"""
BlastHit

The BlastHit object stores all the data related to a blast XML output hit.
    
That means the data inside each <Hit> tag of the blast XML file which corresponds to a single blast match to the query.

WARNING!: It has to be a blastpgp from a single sequence; multi-fasta blasts are not compatible
          with xml.dom.minidom parsing capabilities.

                                   ############################################ 
                                        jbonet @ boliva's lab        2011
"""
import re

from SBI.sequence import IndexedSeqAli as IndexedSeqAli
from collections      import Counter
class BlastHit(IndexedSeqAli):    
    """
    The BlastHit object stores all the data related to a blast XML output hit.
    
    That means the data inside each <Hit> tag of the blast XML file which corresponds to a single blast match to the query.
    
    WARNING!: It has to be a blast from a single sequence; multi-fasta blasts are not compatible
              with xml.dom.minidom parsing capabilities.
    """

    global VALID_ALIGNMENT_SEQ_TYPES
    VALID_ALIGNMENT_SEQ_TYPES = {'query': 0, 0:0, 'hit': 1, 1:1}

    def __init__(self, name, length, iteration = 0,    e_value   = None, align_length      = None,
                 identities = None,  positives = None, gaps      = None, qseq = None, hseq = None, 
                 qpos       = None,  hpos      = None, score_seq = None):
        """
        BlastResult Constructor [PARAMETERS]: [ALL MANDATORY]
            (str)    name           -> name of the hit sequence              -> self.sequenceID
            (int)    length         -> length of the hit sequence            -> self.length
            (int)    iteration      -> iteration to which the hit belongs    -> self.iteration
            (str)    e_value        -> alignment e-value                     -> self.e_value
            (int)    align_length   -> alignment length                      -> self.align_length
            (int)    identities     -> alignment identities                  -> self.identities
            (int)    positives      -> alignment positives                   -> self.positives
            (int)    gaps           -> alignment gaps                        -> self.gaps
            (str)    qseq           -> query sequence aligned                -> self.query_seq
            (str)    hseq           -> hit sequence aligned                  -> self.hit_seq
            (arr)    qpos           -> positions of the aligned query        -> self.qpos[]
            (arr)    hpos           -> positions of the aligned hit          -> self.hpos[]
            
            
            BlastResult [INTERNAL VARIABLES]
            (str)    score_seq      -> score sequence summary
        """        
        super(BlastHit, self).__init__(sequences     = [str(qseq),str(hseq)], 
                                       sequenceInits = [qpos,hpos], 
                                       identities    = identities,
                                       positives     = positives,
                                       gaps          = gaps)

        self.add_alignment_pattern(alipatt = score_seq, method = 'blast')

        self._sequenceID   = str(name)                # Hit Name
        self._length       = int(length)              # Hit Length
        self._iteration    = int(iteration)           # Iteration to which the hit belongs
        
        self._e_value      = e_value                  # Alignment e-value
        self._align_length = int(align_length)        # Alignment Length

        self._used         = False
    
    '''ATTRIBUTES (RENAMED FROM PARENT)'''
    @property
    def query_seq(self):            return str(self._seq[0].sequence)
    @property
    def ungapped_query_seq(self):   return str(re.sub('-','',self.query_seq))
    @property
    def hit_seq(self):              return str(self._seq[1].sequence)
    @property
    def ungapped_hit_seq(self):     return str(re.sub('-','',self.hit_seq))
    @property
    def query_pos(self):            return self._segment[0]
    @property
    def hit_pos(self):              return self._segment[1]

    '''ATTRIBUTES'''
    @property
    def sequenceID(self):           return self._sequenceID
    @property
    def length(self):               return self._length
    @property
    def iteration(self):            return self._iteration
    @property
    def e_value(self):    
        try:          
            return float(self._e_value)
        except:
            return None
    @property
    def align_length(self):         return self._align_length
    @property
    def is_used(self):              return self._used
    @is_used.setter
    def is_used(self, value):       self._used = value

    '''MIX ATTRIBUTES'''
    @property
    def query_threshold_coord(self):        return [self._first_segment_value(0), self._last_segment_value(0)]
    @property
    def query_align_segment_length(self):   return self._aligned_sequence_length(0)
    @property
    def hit_threshold_coord(self):          return [self._first_segment_value(1), self._last_segment_value(1)]
    @property
    def hit_align_segment_length(self):     return self._aligned_sequence_length(1)

    def _get_valid_alignment_seq_types(self):
        global VALID_ALIGNMENT_SEQ_TYPES
        return VALID_ALIGNMENT_SEQ_TYPES

    '''OVERWRITTE PARENT GETTERS TO ADAPT TO refseq=["query","hit"]'''
    def get_respective_coordinate(self, refseq, coordinate):
        refseq = self._check_valid_alignment_seq_type(refseq)
        return super(BlastHit, self).get_respective_coordinate(refseq, int(not refseq), coordinate)

    def get_coverage_of_sequence(self, refseq,  fulllength = None, ini = None, end = None):
        """
        METHOD: get_coverage_of_sequence(refseq =["query","hit", fulllength = None, ini = None, end = None))

            __DEVEL_INFO__
            => Overwrittes from parent (SeqAli)
        
            __EVALUATOR__
            =>Returns coverage ratio (0<->1) of the quey/hit over the full sequence (for query fulllength is required)
            
            =>If a *range* is given, returns __the sequence coverage for the given range__
             (adjusted to actually aligned regions)

            =>[TIP] To obtain __the sequence coverage for the aligned region over itself__:
                ini = self.query_threshold_coord[0] or self.hit_threshold_coord[0]
                AND
                end = self.query_threshold_coord[1] or self.hit_threshold_coord[1]
        """
        refseq = self._check_valid_alignment_seq_type(refseq)
        if bool(refseq): #hit
            fulllength = self.length 
        else:            #query
            if fulllength is None:
                raise AttributeError('To check the query coverage the query length must be given')

        return super(BlastHit, self).get_coverage_of_sequence(VALID_ALIGNMENT_SEQ_TYPES[refseq],fulllength, ini, end)
    
    def get_min_coverage_of_sequence(self, fulllength):
        return min(self.get_coverage_of_sequence('hit'),self.get_coverage_of_sequence('query',fulllength))

    def get_coverage_of_full_sequence_segment(self, refseq):
        refseq = self._check_valid_alignment_seq_type(refseq)
        if bool(refseq): #hit
            ini = self.hit_threshold_coord[0]
            end = self.hit_threshold_coord[1]
            length = len(self.hit_seq.replace('-',''))
        else: #query
            ini = self.query_threshold_coord[0]
            end = self.query_threshold_coord[1]
            length = len(self.query_seq.replace('-',''))

        return super(BlastHit, self).get_coverage_of_sequence(VALID_ALIGNMENT_SEQ_TYPES[refseq],length)

    def get_min_coverage_of_full_sequence_segment(self):
        return min(self.get_coverage_of_full_sequence_segment('hit'),self.get_coverage_of_full_sequence_segment('query'))

    def get_section_from_sequence_position(self, refseq, ini, end):
        refseq = self._check_valid_alignment_seq_type(refseq)
        return super(BlastHit, self).get_section_from_sequence_position(refseq, ini, end)

    def correct_hit_count(self, new_index):
        if isinstance(self._idx[1], list):
            raise AttributeError('Complex index of hit cannot be modified')

        if not isinstance(new_index, list):
            self.increment_sequence_index(1, new_index)
        else:
            self.add_complex_index(1, new_index)

    def correct_query_count(self, new_index):
        if isinstance(self._idx[0], list):
            raise AttributeError('Complex index of querycannot be modified')

        if not isinstance(new_index, list):
            self.increment_sequence_index(0, new_index)
        else:
            self.add_complex_index(0, new_index)
    
    def _check_valid_alignment_seq_type(self, seq_type): 
        """
        METHOD: _check_valid_alignment_seq_type()
        
            __STATUS_CHECKER__
            =>Returns BOOLEAN:
                                True if the seq_type is accepted
                                False otherwise
        """
        if not seq_type in self._get_valid_alignment_seq_types():
            raise ValueError( "'ref_seq' must be in %s. Recieved 'ref_seq' is %s" %(repr(self._get_valid_alignment_seq_types()), seq_type) )
        return self._get_valid_alignment_seq_types()[seq_type]
        
    def overlap(self, blastHit):
        #It returns a range between 0<->1 of how much the actual blastHit and the one passed
        #overlap in the query sequence.
        #The overlap is measured OVER THE SHORTEST lenght
        return super(BlastHit, self).overlap(1, 0, blastHit)
    
    """
        toString
    """
    def __str__(self):
        """
        METHOD: __str__() IS DEFINED
        """

        return "\t".join((self.sequenceID, str(self.length), str(self.identities), str(self.positives), str(self.gaps),
                          "{:.3}".format(self.e_value), self.query_seq, self.hit_seq, self.formatPositions()))

        '''OVEWRITTE OBJECT FUNCTIONS (FROM PARENT)'''
    def __getitem__(self, key):
        # We correct the count as one would try to ask for the first position as 1
        # Runs by alignment position
        # if not self._has_complex_index:
        #     return super(IndexedSeqAli, self).__getitem__(key)
        try:        
            int(key)
            if int(key) > len(self): raise IndexError
            return map(lambda n: n[int(key) - 1],self._seq)
        except:
            if not isinstance(key, slice):  
                raise TypeError()
            else:
                index = key.indices(len(self) + 1)
                if key.stop is not None:
                    index = (index[0],index[1]+1,index[2]) #Remember! we go according to align position, not array!!
                else:
                    index = (index[0],index[1],index[2])
                sequences = ['' for x in range(self.number_of_sequences)]
                seqInits  = ['' for x in range(self.number_of_sequences)]
                complex_indexes = {}
                complex_pregaps = {}
                for i in range(*index):
                    data = self[i]
                    for x in range(len(data)):
                        sequences[x] += data[x]
                    if i == len(self): break

                for x in range(self.number_of_sequences):
                    if not isinstance(self._idx[x], list):
                        seqInits[x] = self._sequence_position_from_alignment_position(x,index[0])
                    else:
                        seqInits[x] = self._idx[x].index(self._sequence_position_from_alignment_position(x,index[0])) + 1
                        complex_indexes[x] = self._idx[x]
                        complex_pregaps[x] = Counter(str(self.sequences[x]).split(str(sequences[x]))[0])['-']
                newali = self.__class__(name = self._sequenceID, length = self._length, iteration = self._iteration,
                                        qseq = sequences[0], hseq = sequences[1], hpos = seqInits[1], qpos = seqInits[0],
                                        align_length = len(sequences[0]), score_seq = self._alipatt[index[0]-1:index[1]-1:index[2]])
                for refident in complex_indexes:
                    newali.add_complex_index(refident, complex_indexes[refident],complex_pregaps[refident])
                if not self._alipatt is None:
                    newali.add_alignment_pattern(self._alipatt[index[0]-1:index[1]-1:index[2]],self._aliptmeth)
                return newali

            


'''
@file: DSSPExe.py

@author: Jaume Bonet
@mail:   jaume.bonet@gmail.com
@date:   2013

@ [oliva's lab](http://sbi.imim.es)

@class: DSSPExe
'''
import os
import sys

from SBI.external import ExternalExe
from DSSP         import DSSP
from SBI          import SBIglobals as SBIg
from SBI.beans    import File


class DSSPExe(ExternalExe):
    '''
    Manages the execution of DSSP, a secondary structure determination program
    from PDB structures.

    '''
    def __init__(self, pdb_file, dssp_file, cleanpdb = False, cleandssp = False):
        '''
        @param:    pdb_file
        @pdef:     name of the PDB file to calculate SS from
        @ptype:    {String}

        @param:    dssp_file
        @pdef:     name of the DSSP output file
        @ptype:    {String}

        @param:    cleanpdb
        @pdef:     flag. delete pdb file after execution
        @pdefault: _False_
        @ptype:    {Boolean}

        @param:    cleandssp
        @pdef:     flag. delete dssp file after execution
        @pdefault: _False_
        @ptype:    {Boolean}
        '''
        self._pdbfile  = pdb_file
        self._dsspfile = dssp_file
        self._dsspdata = []
        self._gapped   = False

        if DSSPExe._EXE is None:
            self._set_default_executable('dssp')

        self._execute()
        self._parse()
        self._clean(cleanpdb, cleandssp)

    ##############
    # ATTRIBUTES #
    ##############
    @property
    def dsspdata(self):
        '''
        List with the DSSP result.
        One residue-data per position.

        @return: {List}
        '''
        return self._dsspdata

    @property
    def empty_dssp(self):
        '''
        Creates a empty DSSP object to fill the gaps.

        @return: {DSSP}
        '''
        return DSSP(secondary_structure = '-',
                    accessibility       = 1,
                    amino               = 'X')

    ############
    # BOOLEANS #
    ############
    @property
    def gapped(self):
        '''
        Checks if dssp has found gaps in the structure.

        @return: {Boolean}
        '''
        return self._gapped

    ###########
    # METHODS #
    ###########
    @staticmethod
    def dynamic(executable, path):
        '''
        Manually set the values for dssp path and executable .

        @param:    executable
        @pdef:     name of the executable file
        @ptype:    {String}

        @param:    path
        @pdef:     path to the executable file
        @ptype:    {String}
        '''
        DSSPExe._set_dynamic_executable(executable, path)

    ###################
    # PRIVATE METHODS #
    ###################
    def _execute(self):
        '''
        Executes the DSSP call.
        '''
        self._EXE.add_parameter(self._pdbfile)
        self._EXE.add_parameter(self._dsspfile)
        try:
            self._EXE.execute(silent=True)
        except SystemError, e:
            msg = 'Some error occurred while executing dssp\n{0}\n'.format(e)
            SBIg.throw(self, msg, e)

        self._EXE.clean_command()

    def _parse(self):
        file_fd    = File(self._dsspfile)
        read       = False
        continuity = -1000
        readline   = 0
        for line in file_fd.read():
            if line.startswith("  #  RESIDUE AA STRUCTURE BP1 BP2  ACC"):
                read = True
                continue
            if read:
                if line[13:14] != '!':
                    res_num = int(line[6:10].strip())
                    ss      = line[16:17] if line[16:17] != ' ' else '-'
                    buried  = int(line[35:38].strip())
                    aa      = line[13:15].strip()

                    self._dsspdata.append(DSSP(secondary_structure = ss,
                                               accessibility       = buried,
                                               amino               = aa))
                    self._dsspdata[-1].add_hydrogen_links(line[39:50],
                                                          line[50:61],
                                                          line[61:72],
                                                          line[72:84])
                    if readline > 0:
                        if res_num != continuity + 1:
                            self._gapped = True
                        continuity = res_num
                    readline += 1
                else:
                    msg = "truncated chain!{0}\n".format(self._dsspfile)
                    sys.stderr.write(msg)
                    SBIg.warn(self, msg)
                    self._gapped = True
        file_fd.close()

    def _clean(self, cleanpdb, cleandssp):
        '''
        If required, removes the files used in the DSSP execution.

        @param:    cleanpdb
        @pdef:     flag. delete pdb file after execution
        @ptype:    {Boolean}

        @param:    cleandssp
        @pdef:     flag. delete dssp file after execution
        @ptype:    {Boolean}
        '''
        if cleanpdb:
            os.unlink(self._pdbfile)
        if cleandssp:
            os.unlink(self._dsspfile)



# -*-
#
# @author: jaumebonet
# @email:  jaume.bonet@gmail.com
# @url:    jaumebonet.github.io
#
# @date:   2015-05-10 16:35:48
# @lab:    LPDI/EPFL
#
# @last modified by:   jaumebonet
# @last modified time: 2015-05-28 10:36:34
#
# -*-
'''
Given a PIR alignment and a minimal set of parameters generates models
through homology modeling with MODELLER.

In this instance, it takes a special care into optimize loop regions.

[!] The script does not contain any input check.
[!] The loop_model function redirects STDOUT and STDERR to file logs.
    The redirection is terminated before the function ends.

'''
import sys
import os
import warnings
from argparse import ArgumentParser

from modpyHelper import basic_parser
from modpyHelper import identify_pir

from modeller import *
from modeller.automodel import *


def set_options(*args, **kargs):
    '''
    Set the specific options for this script

    @return: parsed ArgumentParser
    '''
    parser = ArgumentParser(parents = [basic_parser()],
                            conflict_handler='resolve')

    parser.add_argument("--models", dest="numMod",     action="store",
                        type=int,   metavar="MOD_NUM", default=5,
                        help="Number of models to be done (min: 1) [Default: 5]")
    parser.add_argument("--loops", dest="numLoop",     action="store",
                        type=int,  metavar="LOOP_NUM", default=5,
                        help="Number of loop to be done per model (min: 1) [Default: 5]")
    parser.add_argument("--dopeloop",  dest="dopel", action="store_true",
                        default=False, help="Optimize loops with DOPE")
    parser.add_argument("--optimize",  dest="optimize", action="store_true",
                        default=False, help="Activate optimization")
    return parser.parse_args()


def loop_model(alignment, instances, linstances,
               output=None, dopeloop=False, optimize=False, verbose=False):

    contents = identify_pir(alignment)

    # We redirect STDOUT and STDERR to output files
    if output is None: output = contents['seq']
    warnings.warn('STDOUT & STDERR will be redirected to log files.')
    sys.stdout = open(output + '.log', 'w')
    sys.stderr = open(output + '.err', 'w')

    if verbose: log.verbose()  # Commands MODELLER to display all log output

    env = environ()  # Initializes the 'environment' for this modeling run

    env.io.hetatm = True  # Allow the presence of hetatoms

    if not dopeloop: loopBaseClass = loopmodel
    else:            loopBaseClass = dopehr_loopmodel

    class MyLoop(loopBaseClass):
        pass

    a = MyLoop(env,  # Loading the environment
               alnfile=alignment,         # Assigning the PIR alignment
               knowns=contents['str'],    # Listing the known structures
               sequence=contents['seq'],  # Identify the Query Sequence
               assess_methods=(assess.DOPE,
                               assess.GA341),
               loop_assess_methods=assess.DOPE)  # Energy evaluation methods

    # Setting Starting and Ending Model number
    a.starting_model, a.ending_model = 1, instances
    if linstances <= 0: linstances = 1
    # Setting Starting and Ending Loop number per model
    a.loop.starting_model, a.loop.ending_model = 1, linstances

    if optimize:
        # Very thorough VTFM optimization:
        a.library_schedule   = autosched.slow
        a.max_var_iterations = 300

        # Thorough MD optimization:
        a.md_level      = refine.slow
        a.loop.md_level = refine.slow

        # Repeat the whole cycle 2 times and do not stop unless obj.func. > 1E6
        a.repeat_optimization = 2
        a.max_molpdf          = 1e6

    a.initial_malign3d = True

    a.make()  # Create the Model

    sys.stdout = sys.__stdout__  # restore stdout back to normal
    sys.stderr = sys.__stderr__  # restore stdout back to normal
    warnings.warn('STDOUT & STDERR have been restored.')

    # We delete the error file if no error has occurred
    ERRsize = os.path.getsize(output + '.err')
    if int(ERRsize) is 0:
        os.remove(output + '.err')

if __name__ == '__main__':
    options = set_options()
    loop_model(options.alignment, options.numMod, options.numLoop, options.out,
               options.dopel, options.optimize, options.verbose)



"""
PDB

author: jbonet
date:   02/2013

@oliva's lab
"""

import numpy

from .chain import ChainOfProtein, ChainOfNucleotide

from SBI.beans.StorableObject import StorableObject
from SBI.beans.File           import File


class PDB(StorableObject):

    """
    A {PDB} is a collection of {Chain}
    """
    def __init__(self, pdb_file=None, dehydrate=False, header=False,
                 onlyheader=False, biomolecule=False):
        """
        @type  pdb_file: String
        @param pdb_file: PDB formated file to read

        @raise IOError if pdb_file does not exist and it is not an empty object
        """
        if biomolecule or onlyheader:
            header = True

        self._pdb_file      = pdb_file
        self._chains        = []
        self._NMR           = False
        self._NMR_chains    = []
        self._chain_id      = set()

        self._biomol_id     = -1    # -1 -> original
                                    #  0 -> symmetry
                                    # >0 -> biomolecule

        self._header        = None

        self._has_prot      = False
        self._has_nucl      = False

        self._COMPND        = None

        if self.pdb_file is not None:
            self._pdb_file  = File(file_name=self._pdb_file, action='r')
            self._read_PDB_file(header=header,
                                onlyheader=onlyheader,
                                biomolecule=biomolecule)

        if dehydrate:
            self.dehydrate()

    #
    # ATTRIBUTES
    #
    @property
    def pdb_file(self):
        """
        PDB file name
        @rtype: String
        """
        return self._pdb_file

    @pdb_file.setter
    def pdb_file(self, value):
        """
        Sets a PDB file if none has been given
        @raise UsedAttributeError
        """
        if self._pdb_file is not None:
            raise AttributeError(
                "The PDB object is loaded from file {0}. To load the new file {1} create a new PDB object".format(self._pdb_file.full, value))

        if isinstance(value, File):
            self._pdb_file = value
        else:
            self._pdb_file = File(file_name=value, type='r')

    @property
    def chain_identifiers(self):
        return self._chain_id

    @property
    def id(self):
        return self._chains[0].pdb

    @property
    def chains(self):
        """
        List of {Chain} contained in the PDB w/out NMR replicas
        @rtype: List of {Chain}
        """
        return self._chains

    @property
    def proteins(self):
        """
        List of {ProteinChain} contained in the PDB w/out NMR replicas
        @rtype: List of {ProteinChain} (iterator)
        """
        for chain in self.chains:
            if isinstance(chain, ChainOfProtein):
                yield chain

    @property
    def nucleotides(self):
        """
        List of {NucleotideChain} contained in the PDB w/out NMR replicas
        @rtype: List of {NucleotideChain} (iterator)
        """
        for chain in self.chains:
            if isinstance(chain, ChainOfNucleotide):
                yield chain

    @property
    def non_standard_chains(self):
        """
        List of non {NucleotideChain}/ non {ProteinChain} contained in the PDB w/out NMR replicas
        @rtype: List of non {NucleotideChain}/ non {ProteinChain} (iterator)
        """
        for chain in self.chains:
            if not isinstance(chain, ChainOfNucleotide) and not isinstance(chain, ChainOfProtein):
                yield chain

    @property
    def all_models(self):
        """
        List of {Chain} contained in the PDB w/ NMR replicas
        @rtype: List of {Chain}
        """
        return self._chains + self._NMR_chains

    @property
    def header(self):
        if self._header is None:
            return ''
        else:
            return self._header

    @property
    def biomolecule_identifier(self):
        return self._biomol_id

    #
    # COMPLEX GETTERS & SETTERS
    #
    def get_chain_by_id(self, id):
        """
        Returns a chain according to its id or None if no chain with that id is found
        @rtype: {Chain}
        """
        for chain in self._chains:
            if chain.chain == id:
                return chain
        return None

    def add_chain(self, chain, NMR=False):
        """
        Adds a new chain to the PDB
        """
        if not NMR:
            self._chains.append(chain)
        elif NMR and self._NMR:
            self._NMR_chains.append(chain)

        self._chain_id.add(chain.chain)

    def add_chains(self, chains, NMR=False):
        """
        Adds a new chains to the PDB
        """
        for chain in chains:
            self.add_chain(chain=chain, NMR=NMR)

    def _get_chain_position_by_id(self, id):
        """
        Returns the position in the chain array where the chain is
        @rtype: Integer
        """
        for x in range(len(self._chains)):
            if self._chains[x].chain == id:
                return x
        return None

    #
    # BOOLEANS
    #
    @property
    def is_NMR(self):
        """
        Identifies if the PDB contains NMRs
        @rtype: Boolean
        """
        return self._NMR

    def chain_exists(self, chain):
        """
        Confirms if a given chain exists in the PDB
        @rtype: Boolean
        """
        return chain in self._chain_id

    @property
    def has_protein(self):
        """
        Checks if the PDB contains a protein (not only)
        @rtype: Boolean
        """
        return self._has_prot

    @property
    def has_nucleotide(self):
        """
        Checks if the PDB contains a nucleotide chain (not only)
        @rtype: Boolean
        """
        return self._has_nucl

    @property
    def repeated_chain_ids(self):
        """
        Checks if more than one {Chain} has the same assigned ID
        @rtype: Boolean
        """
        return len(self._chain_id) < len(self._chains)

    @property
    def is_all_ca(self):
        for p in self.proteins:
            if p.is_only_ca():
                return True
        return False

    #
    # METHODS
    #
    def dehydrate(self):
        recheck_chains = False
        for c in self.chains:
            c.dehydrate()
            if c.is_empty:
                recheck_chains = True
        if recheck_chains:
            c = []
            for ch in self.chains:
                if not ch.is_empty:
                    c.append(ch)
                else:
                    self._chain_id.remove(ch.chain)
            self._chains = c

    def duplicate(self, hetero=True, water=False, NMR=False):
        """
        Returns a {PDB} identical to the original but as a new object
        @rtype: {PDB}
        """
        new_PDB = PDB()
        new_PDB.pdb_file = self.pdb_file

        for chain in self.chains:
            new_PDB.add_chain(
                chain=chain.duplicate(hetero=hetero, water=water))

        if NMR:
            for chain in self._NMR_chains:
                new_PDB.add_chain(chain=chain.duplicate(
                    hetero=hetero, water=water), NMR=True)

        new_PDB._NMR = self._NMR
        new_PDB._has_prot = self._has_prot
        new_PDB._has_nucl = self._has_nucl

        return new_PDB

    def apply_symmetry_matrices(self):
        """
        Only works if the PDB file is an original PDB file
        or the matrices have been added in the correct PDB format
        @rtype: {PDB}
        """
        if self._header is None:
            self._read_PDB_file(header=True, onlyheader=True)
        return self._apply_matrix(matrix=self.header.symmetry_matrix)

    def apply_biomolecule_matrices(self, keepchains=False, water=True):
        """
        Only works if the PDB file is an original PDB file or
        the matrices have been added in the correct PDB format
        @rtype: {PDB}
        """
        if self._header is None:
            self._read_PDB_file(header=True, onlyheader=True)
        PDB_list = []
        for matrix in self.header.biomolecules:
            PDB_list.append(self._apply_matrix(matrix=matrix,
                                               keepchains=keepchains,
                                               water=water))
        return PDB_list

    def _apply_matrix(self, matrix, keepchains=False, water=True):
        new_PDB            = PDB()
        new_PDB._biomol_id = matrix.identifier

        for chain in self.chains:
            if chain.chain in matrix.chains:
                for mat in matrix.matrices:
                    new_chain = chain.duplicate(water=water)
                    new_chain.reposition(matrix=mat.matrix, vector=mat.vector)
                    if len(new_chain) >= 1:
                        new_PDB.add_chain(chain=new_chain)
        if not keepchains:
                new_PDB.tmpclean(cluster_by_alternative_id=True)
        return new_PDB

    def clean(self):
        first_atom = 1
        for c in self.chains:
            c.clean(initatom=first_atom)
            first_atom = c.last_residue.last_atom_number + 1

    def tmpclean(self, cluster_by_alternative_id=False):
        """
        Makes a clean version of the PDB, rechaining in order and renumerating atoms.
        Renumbering residues is optional
        """
        pchainsIDs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890"
        chainsIDs = ""
        chainsNIDs = ""
        chainID = 0
        atom_count = 1

        for x in range(len(pchainsIDs)):
            if not self.chain_exists(chain=pchainsIDs[x]):
                chainsIDs += pchainsIDs[x]
            else:
                chainsNIDs += pchainsIDs[x]

        chain_change = len(self) <= len(chainsIDs)

        for chain in self.chains:
            if (not chain.chain in chainsNIDs) and chain_change:
                self._chain_id.add(chain.chain)
                chain.chain = chainsIDs[chainID]
                chainID += 1
                self._chain_id.add(chain.chain)
                if cluster_by_alternative_id:
                    if self._COMPND is None:
                        self._COMPND = {}
                    if not self._COMPND.has_key(chain.alternative_id):
                        self._COMPND.setdefault(
                            chain.alternative_id, []).append(chain.alternative_id)
                    self._COMPND[chain.alternative_id].append(chain.chain)
            else:
                chainsNIDs = chainsNIDs.replace(chain.chain, '')

            chain.renumerate_atoms(init=atom_count)
            atom_count += (chain.atom_length)

    def fuse_chains(self, chains_ids):
        """
        Fuses several chains into the first one.
        It will not allow to fuse different structural chains.
        It does not alter the {PDB}, but provides a new one
        @rtype: {Chain}

        @raise AttributeError if:
            a) A given chain ID is not present
            b) Try to fuse different structural chains
        """
        if len(self._chain_id.intersection(set(chains_ids))) < len(chains_ids):
            raise AttributeError(
                "Some of the given chains to fues do not exist")

        error_counter = 0
        error_control = [False, False]
        new_PDB = PDB()
        for c in chains_ids:
            chain = self.get_chain_by_id(id=c)
            new_PDB.add_chain(chain=chain.duplicate())
            if isinstance(chain, ChainOfProtein) and not error_control[0]:
                error_counter += 1
                error_control[0] = True
            elif isinstance(chain, ChainOfNucleotide) and not error_control[1]:
                error_counter += 1
                error_control[1] = True
            if error_counter == 2:
                raise AttributeError(
                    "Fuse different kinds of structural chain is not possible\n")

        init_chain_num = new_PDB.chains[0].last_residue.number
        for x in range(1, len(new_PDB.chains)):
            new_PDB.chains[x].renumerate_residues(init=init_chain_num + 1)
            init_chain_num = new_PDB.chains[0].last_residue.number
            new_PDB.chains[0].fuse(chain=new_PDB.chains[x])

        return_PDB = PDB()
        return_PDB.add_chain(chain=new_PDB.chains[0])
        return return_PDB

    # def calculate_dssp(self, out_dir = None, store = True):
    #     """
    #     Executes DSSP and assigns the prediction to each chain

    #     @param  out_dir: directory to save the output
    #     @defaut out_dir: None

    #     @param store: Save the dssp output(?)
    #     """

    #     for chain in self.proteins:
    #         if out_dir is None:
    #             pdb_file  = chain.globalID + ".pdb2dssp"
    #             dssp_file = chain.globalID + ".dssp"
    #         else:
    #             Path.mkdir(newdir = out_dir)
    #             pdb_file  = os.path.join(os.path.abspath(out_dir), chain.globalID + ".pdb2dssp")
    #             dssp_file = os.path.join(os.path.abspath(out_dir), chain.globalID + ".dssp")

    #         pdb_fd = open(pdb_file, 'w')
    #         pdb_fd.write(chain.PDB_format())
    #         pdb_fd.close()

    #         dssp_calc = DSSPexec(pdb_file = pdb_file, dssp_file = dssp_file,
    #                              chain    = chain,    store     = store)

    def rotate(self, matrix=None):
        """
        Rotates each {Chain} according to a given matrix

        @type matrix: numpy.matrix
        """
        if matrix is None:
            matrix = numpy.identity(3, float)
        for chain in self.all_models:
            chain.rotate(matrix=matrix)

    def translate(self, vector=None):
        """
        Translates each {Chain} according to a translational vector

        @type vector: numpy.array
        """
        if vector is None:
            vector = numpy.zeros(3, float)
        for chain in self.all_models:
            chain.translate(vector=vector)

    def reposition(self, matrix=None, vector=None):
        """
        Rotates and Translates each {Chain} according to a matrix and a translational vector

        @type matrix: numpy.matrix

        @type vector: numpy.array
        """
        if matrix is None:
            matrix = numpy.identity(3, float)
        if vector is None:
            vector = numpy.zeros(3, float)
        for chain in self.all_models:
            chain.reposition(matrix=matrix, vector=vector)

    # def calculate_protein_heteroatom_contacts(self, distance = 6):
    #     """
    #     Returns a {HeteroatomContacts} list with the contacts between a protein and its heteroatoms
    #     at a maximum given distance
    #     @type distance: Integer
    #     @rtype: list of {HeteroatomContacts}
    #     """
    #     data = []
    #     for protein in self.proteins:
    #         data.append(HeteroatomContacts(chain = protein, max_distance = distance))
    #     return data

    #
    # OVERRIDE PARENT'S FUNCTIONS
    #
    @staticmethod
    def read(input_file, format='PDB'):
        """
        Reads a file of data in a specific format and returns the object

        @type  input_file: String
        @param input_file: File to read

        @type  format: String
        @param format: Format of the file to read
        """
        if format == 'PDB':
            pdb = PDB(pdb_file=input_file)
            return pdb

    def write(self, output_file=None, format='PDB', force=False, clean=False):
        """
        Writes the object in a specific format

        @type  output_file: String
        @param output_file: File to write

        @type  format: String
        @param format: Format of the file to print
        """
        outfile = File(file_name=output_file, action='w', overwrite=force)
        if format == 'PDB':
            self._write_PDB_file(pdb_file=outfile, clean=clean)

    #
    # IO
    #
    def _read_PDB_file(self, header=False, onlyheader=False, biomolecule=False):
        """
        Process and load crystal data from a PDB formated file
        """
        from parse_pdb import read_PDB_file, read_PDB_header
        if header:
            read_PDB_header(self)
            self._pdb_file.close()
        if not onlyheader:
            # read_PDB_file(self, biomolecule=biomolecule)
            read_PDB_file(self)
        self._pdb_file.close()

    # def _represent_COMPND(self):
    #     if self._COMPND is None: return ''

    #     data = []
    #     mol_counter = 1
    #     for chain in self._COMPND:
    #         data.append("COMPND    MOL_ID: %d;" %mol_counter)
    #         data.append("COMPND   2 CHAIN: " + ",".join(self._COMPND[chain]) + ";")
    #         if len(self._biomolecA) > 0:
    #             matrices = []
    #             for mat in self._biomolecA:
    #                 if mat[1] == chain: matrices.append(str(mat[0]))
    #             data.append("COMPND   3 MATRICES: " + ",".join(sorted(matrices)))
    #         mol_counter += 1
    #     return "\n".join(data) + "\n"

    def _write_PDB_file(self, pdb_file, clean=False):
        """
        Print a crystal into a PDB formated file
        """
        out_fd = pdb_file.descriptor
        # out_fd.write(self._represent_COMPND())
        out_fd.write(self.PDB_format(clean=clean) + "\n")
        pdb_file.close()

    def PDB_format(self, clean=False, terminal=True):
        """
        Strings a {PDB} in PDB format
        @rtype: String
        """
        lines = []
        if clean:
            self.clean()
        for chain in self._chains:
            lines.append(chain.PDB_format(terminal=terminal))
        lines.append("END")

        return "\n".join(lines)

    def FASTA_format(self, gapped=True, protein=True, nucleotide=False):
        lines = []
        for c in self.chains:
            if isinstance(c, ChainOfProtein) and protein:
                lines.append(
                    ">{0}\t{1}".format(c.globalID, c.aminoacids[0].identifier))
                if gapped:
                    lines.append("{0}".format(c.gapped_protein_sequence))
                else:
                    lines.append("{0}".format(c.protein_sequence))
            if isinstance(c, ChainOfNucleotide) and nucleotide:
                lines.append(
                    ">{0}\t{1}".format(c.globalID, c.nucleotides[0].identifier))
                if gapped:
                    lines.append("{0}".format(c.gapped_nucleotide_sequence()))
                else:
                    lines.append("{0}".format(c.nucleotide_sequence()))
        if len(lines) == 0:
            return ""
        else:
            return "\n".join(lines) + "\n"

    def IDX_format(self, protein=True, nucleotide=False):
        lines = []
        for c in self.chains:
            if isinstance(c, ChainOfProtein) and protein:
                lines.append(">{0}\t{1}".format(c.globalID, c.protein_idx))
            if isinstance(c, ChainOfNucleotide) and nucleotide:
                lines.append(
                    ">{0}\t{1}".format(c.globalID, c.nucleotide_idx()))
        if len(lines) == 0:
            return ""
        else:
            return "\n".join(lines) + "\n"

    def FASTA_IDX(self, protein=True, nucleotide=False):
        data = {}
        data.setdefault('FASTA', [])
        data.setdefault('IDX', [])
        for c in self.chains:
            if isinstance(c, ChainOfProtein) and protein:
                data['FASTA'].append(
                    ">{0}\n{1}".format(c.globalID, c.gapped_protein_sequence))
                data['IDX'].append(
                    ">{0}\t{1}".format(c.globalID, c.protein_idx))
            if isinstance(c, ChainOfNucleotide) and nucleotide:
                data['FASTA'].append(
                    ">{0}\n{1}".format(c.globalID, c.gapped_nucleotide_sequence()))
                data['IDX'].append(
                    ">{0}\t{1}".format(c.globalID, c.nucleotide_idx()))

        return data

    #
    # OVERRIDE DEFAULT METHODS
    #
    def __len__(self):
        return len(self._chains)



import sys, os, subprocess

class LocalSubmitter(object):
	'''
	Submit jobs to Local computer
	'''
	def __init__(self, log_path='./'):
		self.log_path = log_path

	def submit(self, cmd, job_id, verbose=True):
		'''
		Submits a job to Local computer
		'''
		if verbose: sys.stdout.write('Processing %s...\n' % job_id)
		CLI = '%s > %s.out 2> %s.err' % (cmd, os.path.join(self.log_path, job_id) , os.path.join(self.log_path, job_id))
		subprocess.call(CLI, shell=True)

class GaudiSubmitter(object):
	'''
	Submit jobs to Gaudi Cluster
	'''
	def __init__(self, qsub='sbi', log_path='./'):
		self.qsub = qsub
		self.log_path = log_path

	def submit(self, cmd, job_id):
		'''
		Submits a job to Gaudi cluster
		Cluster params:
		-q   : Bind job to queue
		-cwd : Use current working directory
		-V   : Export all environmental variables
		-e   : Standard error stream log_path
		-o   : Standard out stream path
		-N   : Job name
		'''
		CLI = 'echo "%s" | qsub -q %s -cwd -V -o %s -e %s -N %s' % (cmd, self.qsub, self.log_path, self.log_path, job_id)
		subprocess.call(CLI, shell=True)


import sys
import argparse
import os
import re
from Bio import SeqIO as SeqIE
from Bio import ExPASy
from Bio.Seq import Seq
from Bio import SeqRecord
from Bio.Alphabet import IUPAC
import ConfigParser
import itertools
import shutil
import subprocess
import time
import random


# Add '.' to sys.path
src_path = os.path.abspath(os.path.dirname(__file__))
sys.path.append(src_path)

# Read configuration file
config = ConfigParser.ConfigParser()
config_file = os.path.join(src_path, 'config.ini')
config.read(config_file)

# Add SBI library path to sys.path
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','sbi_library_path'))
sys.path.append(src_path)

# Add functions path to sys.path
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','functions_path'))
sys.path.append(src_path)

# Import functions
import functions
from SeqIO import *
from functions import *
from SBI.external.blast import blast_parser
from SBI.structure.contacts import Complex
from SBI.sequence import Sequence
from SBI.structure import PDB

class ModelException(Exception):
    pass


def main():

    options    = parse_user_arguments()
    verbose    = options.show
    dummy_dir  = options.dummy_dir
    if dummy_dir.startswith("/"):
     cluster_dummy_dir=dummy_dir
    else:
     cluster_dummy_dir=os.path.abspath(dummy_dir)
    outdir     = options.outdir
    parallel   = options.parallel
    query_list = options.query_list
    nmodels    = int(options.nmodels)
    optimize   = options.optimize
    did        = options.did
    fasta      = options.sequence_file
    analysis   = options.analysis
    label      = options.label
    cont       = options.cont
    hydrogens  = options.hbplus
    renumerate = options.renumerate
    zrank      = options.zrank
    foldx      = options.foldx
    rosetta    = options.rosetta
    ssp        = options.ssp

    if label is None:
       label=query_list.split("/")[-1]

    if not os.path.exists(outdir): os.makedirs(outdir)
    make_subdirs(outdir, subdirs =['models'])
    modeldir=os.path.join(outdir,'models')

    flags = " -skip -n %d "%(nmodels)
    energy = " "
    if optimize:  flags = flags + " -opt "
    if did:       flags = flags + " -3did "
    if verbose:   flags = flags + " -v "
    if hydrogens: flags = flags + " --hydrogens "
    if renumerate:flags = flags + " --renumerate "
    if zrank:     energy = energy + " --zrank"
    if foldx:     energy = energy + " --foldx"
    if ssp:       energy = energy + " --split-potentials"
    if rosetta:   energy = energy + " --rosetta"   
    translate={}
    if fileExist(fasta):
     for protein in FASTA_iterator(fasta):
        name = protein.get_identifier()
        entry=accession=name
        if re.search('[a-z][|A-Z0-9][|A-Z0-9]', name):
         if len(name.split("|"))>2:
           accession=name.split("|")[1] 
           entry    =name.split("|")[2] 
        translate.setdefault(entry.split()[0],accession.split()[0])
    else:
        sys.stderr.write('EXIT: Missing FASTA sequence list file %s\n' %(fasta))
        exit(0)


    fd=open(query_list,"r")
    pair={}
    for line in fd:
       if line.startswith("#"): continue
       try:
        p,q=line.strip().split()
       except:
        continue
       try:
        a=translate[p]
        b=translate[q]
       except:
        a=p
        b=q
       x=a.split("_")[0]
       y=b.split("_")[0]
       if pair.has_key((x,y)) or pair.has_key((y,x)):
        if pair.has_key((x,y)):
         pair.setdefault((x,y),set()).add((a,b))
        else:
         pair.setdefault((y,x),set()).add((b,a))
       else:
         pair.setdefault((x,y),set()).add((a,b))
         pair.setdefault((x,y),set()).add((x,y))
    fd.close()
  
    
    for wt,forms in pair.iteritems():
     output=open(os.path.join(outdir,label+"_"+wt[0]+"_"+wt[1]+".ppi"),"w")
     for p,q in forms:
      output.write("%s\t%s\n"%(p,q))
     output.close()


    src_path    = config.get('Paths','modppi_path')   
    scripts_path= os.path.join(src_path,'scripts')
    python_path = config.get('Paths', 'python_path')

    for wt,forms in pair.iteritems():
      ppi=os.path.join(outdir,label+"_"+wt[0]+"_"+wt[1]+".ppi")
      if parallel:
       if  config.get("Cluster", "cluster_queue") == "None": cluster_queue=None
       else: cluster_queue=config.get("Cluster", "cluster_queue")
       functions.submit_command_to_queue("%s %s %s -seq %s -ppi %s -o %s -d %s" % ( os.path.join(python_path, "python"),  os.path.join(scripts_path,'modppi.py'), flags, os.path.abspath(fasta), os.path.abspath(ppi), os.path.abspath(modeldir), cluster_dummy_dir),cluster_queue, int(config.get("Cluster", "max_jobs_in_queue")),os.path.join(scripts_path,config.get("Cluster","command_queue")),cluster_dummy_dir,config.get("Cluster","cluster_submit"),config.get("Cluster","cluster_qstat"))
      else:
       os.system("%s %s %s -seq %s -ppi %s -o %s -d %s" % ( os.path.join(python_path, "python"),  os.path.join(scripts_path,'modppi.py'), flags, os.path.abspath(fasta), os.path.abspath(ppi), os.path.abspath(modeldir),dummy_dir))
 
    if parallel and analysis and not cont:
      sys.stderr.write("Wait until all submissions have finished, then run again with flag '--continue'\n")
      exit(0)  
      
    if analysis:
     pair_model={}
     interactions_done=open(os.path.join(modeldir,'interactions_done.list'),"r")
     for line in interactions_done:
       word=line.strip().split()
       has_model=word[2]
       if has_model=="FAILED":continue
       a,b,has_model,path_list=line.strip().split()
       if pair_model.has_key((a,b)):
          pair_model.pop((a,b))
          pair_model.setdefault((a,b),path_list)
       else:
          pair_model.setdefault((a,b),path_list)
     interactions_done.close()
     for wt,forms in pair.iteritems():
      done_list =[(p,q) for p,q in forms if pair_model.has_key((p,q)) or  pair_model.has_key((q,p)) ]
      study_list=open(os.path.join(outdir,label+"_"+wt[0]+"_"+wt[1]+".list"),"w")
      for (p,q) in done_list:
        if pair_model.has_key((p,q)):
          study_list.write('%s\t%s\tDONE\t%s\n'%(p,q,pair_model[(p,q)]))
        else:
          study_list.write('%s\t%s\tDONE\t%s\n'%(q,p,pair_model[(q,p)]))
      study_list.close()
      if parallel:
       if  config.get("Cluster", "cluster_queue") == "None": cluster_queue=None
       else: cluster_queue=config.get("Cluster", "cluster_queue")
       functions.submit_command_to_queue("%s %s -l %s -ppi %s --hydrogens --renumerate  -boxplot -v -d %s -o %s -seq %s %s " % ( os.path.join(python_path, "python"),  os.path.join(scripts_path,'analysis.py'),   label+"_"+wt[0]+"_"+wt[1], os.path.abspath(os.path.join(outdir,label+"_"+wt[0]+"_"+wt[1]+".list")), cluster_dummy_dir, os.path.abspath(outdir),os.path.abspath(fasta),energy ), cluster_queue, int(config.get("Cluster", "max_jobs_in_queue")),os.path.join(scripts_path,config.get("Cluster","command_queue")),cluster_dummy_dir,config.get("Cluster","cluster_submit"),config.get("Cluster","cluster_qstat"))
      else:  
       os.system("%s %s  -l %s -ppi %s --hydrogens --renumerate  -boxplot -v -d %s -o %s -seq %s %s " % ( os.path.join(python_path, "python"),  os.path.join(scripts_path,'analysis.py'),   label+"_"+wt[0]+"_"+wt[1], os.path.abspath(os.path.join(outdir,label+"_"+wt[0]+"_"+wt[1]+".list")), dummy_dir,os.path.abspath(outdir),os.path.abspath(fasta),energy)) 
 

      
    sys.stdout.write("All requested submissions are done.\n")
      
        
       
     
def parse_user_arguments(*args, **kwds):
    parser = argparse.ArgumentParser(
        description = 'Automatic modelling of protein-protein interactions',
        epilog      = '@Oliva\'s lab 2016')
    parser.add_argument('-i', '--query_list', dest = 'query_list', action = 'store',
                        help = 'Input file with a list of pairs of proteins to group and test')
    parser.add_argument('-l', '--label_name', dest = 'label', action = 'store', default = None,
                        help = 'Label to store the analyses (default is the name of the input query)')
    parser.add_argument('-seq', '--sequences', dest = 'sequence_file', action = 'store',
                        help = 'Input file with a list of sequences in FASTA format')
    parser.add_argument('-o', '--output_directory', dest = 'outdir', action = 'store', default = 'ModPPI',
                        help = 'Output directory (default is ModPPI)')
    parser.add_argument('-n', '--number_of_models', dest = 'nmodels', action = 'store', default = 1, type=int,
                        help = 'Number of models for each template (default is 1)')
    parser.add_argument('-d', '--dummy_dir', dest = 'dummy_dir', action = 'store', default = '/tmp/modppi_dummy',
                        help = 'Specifies the location of the dummy folder (default is /tmp/modppi_dummy)')
    parser.add_argument('-opt', '--optimize', dest = 'optimize', action = 'store_true',
                        help = 'Flag to allow model optimization (default is False)')
    parser.add_argument('-3did', '--use_domain_interactions', dest = 'did', action = 'store_true',
                        help = 'Flag to include domain-domain interactions from 3DiD (default is False)')
    parser.add_argument('-a','--analysis', dest = 'analysis', action = 'store_true',
                        help = 'Flag to include the analysis of ddG in the runs (default is False)')
    parser.add_argument('-c','--continue', dest = 'cont', action = 'store_true',
                        help = 'Flag to continue with analysis after all models are done in a cluster (default is False)')
    parser.add_argument('-v', '--verbose', dest = 'show', action = 'store_true',
                        help = 'Flag for verbose mode (default is False)')
    parser.add_argument("-j","--parallel", default=False, action="store_true", dest="parallel", 
                        help="Submit JOBS to Queues in parallel (default = False)")
    parser.add_argument('-hydro','--hydrogens', dest = 'hbplus', action = 'store_true',
                        help = 'Flag to include hydrogens in modelling (default is False, and always True for analyses)')
    parser.add_argument('-r','--renumerate', dest = 'renumerate' , action = 'store_true',
                        help = 'Flag to renumber the sequences as in the original FastA (default is False, and always True for analyses)')
    parser.add_argument('-zrank', '--zrank', dest = 'zrank', action = 'store_true',
                        help = 'Flag to calculate energies with ZRANK (default is False)')
    parser.add_argument('-foldx', '--foldx', dest = 'foldx', action = 'store_true',
                        help = 'Flag to calculate energies with FoldX (default is False)')
    parser.add_argument('-rosetta', '--rosetta', dest = 'rosetta', action = 'store_true',
                        help = 'Flag to calculate energies with ROSETTA (default is False)')
    parser.add_argument('-ssp', '--split-potentials', dest = 'ssp', action = 'store_true',
                        help = 'Flag to calculate energies with Split-Statistic Potentials (default is False unless no other method is used)')
    options = parser.parse_args()
    return options


if __name__ == '__main__':
    main()



"""Uniprotlink

author: jbonet
date:   10/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
import sys, os, re
import subprocess
import warnings
import urllib

"""
Dependences in SBI library
"""
from SBI               import SBIglobals
from SBI.databases import Uniprotftp
from SBI.beans.Path     import Path
from SBI.beans.File     import File
from SBI.beans.StorableObject     import StorableObject

class Uniprotlink(object):
    """The Uniprotlink class controls the download and parsing of Uniprot database

    """
    def __init__(self, local = None):
        self._local   = os.path.abspath(local)
        self.__name__ = 'databases.Uniprotlink'    # This must be included in every class for the SBIglobals.alert()
        self._swsfile = 'swisprot.gz'
        self._swsfast = 'swisprot.fasta'
        self._trbfile = 'trembl.gz'
        self._trbfast = 'trembl.fasta'

        if local is not None:
            self.local = local

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):
        self._local   = os.path.abspath(value)
        self._swsfile = os.path.join(self._local, self._swsfile)
        self._swsfast = os.path.join(self._local, self._swsfast)
        self._trbfile = os.path.join(self._local, self._trbfile)
        self._trbfast = os.path.join(self._local, self._trbfast)

    @property
    def localUniprots(self):
        for uni_line in self.localSwissprots:
            yield uni_line
        for uni_line in self.localTrembls:
            yield uni_line

    @property
    def localSwissprots(self):
        swsFile = File(self._swsfile, 'r')
        for uni_line in swsFile.descriptor:
            yield uni_line

    @property
    def localTrembls(self):
        tblFile = File(self._trbfile, 'r')
        for uni_line in tblFile.descriptor:
            yield uni_line

    @property
    def source(self):
        return Uniprotftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def download(self):
        if not self.has_local:
            raise NameError('A local Uniprot database directory must be defined.')

        Path.mkdir(self.local)
        destination = os.path.join(self.local, 'uniprot_sprot.dat.gz')
        urllib.urlretrieve(Uniprotftp['swissprot'], destination)
        destination = os.path.join(self.local, 'uniprot_trembl.dat.gz')
        urllib.urlretrieve(Uniprotftp['trembl'], destination)

        self._process()

        return True

    def get_Uniprot(self, Uniprot):
        if self.has_local:
            for uni_line in self.localUniprots:
                if uni_line.split('\t')[0] == Uniprot:
                    return uni_obj
        else:
            raise NameError('A local Uniprot database directory must be defined.')

    def get_Uniprots(self, UNIset):
        if isintance(UNIset, str):
            warnings.warn('For single Uniprot search the get_Uniprot function is recomended.')
            yield self.get_Uniprot(UNIset)

        if self.has_local:
            for uni_line in self.localUniprots:
                if uni_line.split('\t')[0] in UNIset:
                    yield uni_line
        else:
            raise NameError('A local Uniprot database directory must be defined.')

    """PRIVATE METHODS"""
    def _process(self):
        self._parse_uniprot_file(source = os.path.join(self.local, 'uniprot_sprot.dat.gz'),
                                 destination = self._swsfile, fasta = self._swsfast, code = 'swissprot')
        self._parse_uniprot_file(source = os.path.join(self.local, 'uniprot_trembl.dat.gz'),
                                 destination = self._trbfile, fasta = self._trbfast, code = 'trembl')

    def _parse_uniprot_file(self, source, destination, fasta, code):
        sourceFile = File(source, 'r')
        destinFile = File(destination, 'w', True)
        fastaFile  = File(fasta, 'w', True)
        protein = None
        for line in sourceFile.descriptor:
            if line.startswith('ID'):
                protein = Uniprot(line.split()[1].strip(), code)
            if line.startswith('AC'):
                protein.accession = line.split()[1:]
            if line.startswith('OX'):
                protein.taxid     = line.split()[1]
            if line.startswith('OH'):
                protein.hosts     = line.split()[1]
            if line.startswith('DR'):
                protein.databases = line.split()[1:3]
            if line.startswith('  '):
                protein.sequence  = line.strip().replace(' ','')
            if line.startswith('//'):
                destinFile.write(str(protein) + "\n")
                fastaFile.write(repr(protein) + "\n")
        sourceFile.close()
        destinFile.close()

class Uniprot(object):
    def __init__(self, entry = None, source = None, inline = None):
        if inline is not None:
            inline      = inline.strip()
        self.entry      = entry  if inline is None else inline.split('\t')[0]
        self.source     = source if inline is None else inline.split('\t')[1]
        self._accession = []     if inline is None else eval(inline.split('\t')[2])
        self._ox        = None   if inline is None else inline.split('\t')[3] if inline.split('\t')[3] != 'None' else None
        self._oc        = []     if inline is None else eval(inline.split('\t')[4])
        self._db        = []     if inline is None else eval(inline.split('\t')[5])
        self._seq       = None   if inline is None else inline.split('\t')[6] if inline.split('\t')[6] != 'None' else None

    @property
    def accession(self): return self._accession
    @accession.setter
    def accession(self, value):
        self._accession.extend([x.strip(';') for x in value])

    @property
    def taxid(self): return self._ox
    @taxid.setter
    def taxid(self, value):
        if self._ox is None:
            self._ox = self._extract_taxid(value)
        else:
            raise('More than one taxid! {0}'.format(value))

    @property
    def hosts(self): return self._oc
    @hosts.setter
    def hosts(self, value):
        self._oc.append(self._extract_taxid(value))

    @property
    def sequence(self): return self._seq
    @sequence.setter
    def sequence(self, value):
        if self._seq is None:
            self._seq = value.strip()
        else:
            self._seq += value.strip()

    @property
    def databases(self): return self._db
    @databases.setter
    def databases(self, value):
        self._db.append((value[0].strip(';'),value[1].strip(';')))

    def _extract_taxid(self, value):
        import re
        taxid_id = re.compile('NCBI_TaxID\=(\d+);')
        m        = taxid_id.search(value)
        return m.group(1)

    """OVERWRITE PARENT METHODS"""
    def __repr__(self):
        return ">{0.entry}\n{0.sequence}".format(self)

    def __str__(self):
        return "{0.entry}\t{0.source}\t{0.accession}\t{0.taxid}\t{0.hosts}\t{0.databases}\t{0.sequence}".format(self)



'''
@file: BlastExe.py

@author: Jaume Bonet
@mail:   jaume.bonet@gmail.com
@date:   2013

@ [oliva's lab](http://sbi.imim.es)

@class: BlastExe
@class: BlastParser
@class: BlastError
'''
import os
import time
import re
import ConfigParser
from abc import ABCMeta

from bs4 import BeautifulSoup

from SBI.external import ExternalExe
from SBI.beans    import Executable
from SBI.beans    import File
from SBI.beans    import Path
from SBI.sequence import Fasta
from SBI          import SBIglobals as SBIg
from BlastResult  import BlastResult
from BlastResult  import BlastHeader
from BlastHit     import BlastHit


class BlastExe(ExternalExe):
    '''
    Executes [BLAST](http://blast.ncbi.nlm.nih.gov/Blast.cgi).
    See the [user guide](http://www.ncbi.nlm.nih.gov/books/NBK1763/) to learn
    which parameters to add.

    In case a multi fasta is given as database but it is not formatted,
    it will automatically call the formatting of the file.
    '''
    _DBFORMATER = None
    _EXEC_TYPES = {'protein':    ['.phr', '.pin', '.psq'],
                   'nucleotide': ['.nhr', '.nin', '.nsq']}
    _HIT_ID_FRM = frozenset(['single', 'double', 'all'])
    _BLOCKED    = frozenset(['-db', '-outfmt'])

    def __init__(self, database, search_type = 'protein', self_hit = False,
                 hitid_format = 'single', overwrite = None, clean = True):
        '''
        @param:    database
        @pdef:     database to blast upon.
        @ptype:    {String}

        @param:    search_type
        @pdef:     type of database search.
        @pdefault: 'protein'
        @ptype:    {String}

        @@param:   self_hit
        @pdef:     when _True_ if the query is found in the database, it is
                   retrieved.
        @pdefault: _False_
        @ptype:    {Boolean}

        @param:    hitid_format
        @pdef:     format of the name of the hit. If given a wrong option,
                   it defaults to 'single'
        @pdefault: 'single'
        @poptions: 'single' -> first word of the name,
                   'double' -> first two words of the hit name,
                   'all'    -> all the text in the hit name
        @ptype:    {String}

        @param:    overwrite
        @pdef:     For writing actions. Decides whether it can overwrite an
                   existing file.
        @pdefault: _SBIglobals.overwrite_
        @ptype:    {Boolean}

        @param:    clean
        @pdef:     remove the temporary files after the data is read.
        @pdefault: _True_
        @pclash:   if _SBIglobals.debug_ is _True_, clean is _False_
        @ptype:    {Boolean}

        @raises: {BlastError}
        '''
        self._error = BlastError()

        self._search_type = self._check_execution_type(search_type)

        # Blast executable configuration
        if self._EXE is None:
            self._set_default_executable('blast')
            BlastExe._DBFORMATER = BlastExe._CONFIG.get('blast', 'dbformatexe')

        # Database Configuration
        self._idx      = None
        self._database = self._check_database(os.path.abspath(database))

        # Optional execution parameters
        self._parameters = []

        msg  = ['New Blast Executable created.', ]
        msg.append('Blast executable at {0}'.format(self._EXE.full_executable))
        SBIg.alert('debug', self, msg)

        self._self_hit     = self_hit
        self._hitid_format = self._check_hitid_format(hitid_format)
        self._clean_files  = clean

        # Local overwrite takes precedence over Global overwrite
        self._overwrite = SBIg.decide_overwrite(overwrite)

    ##############
    # ATTRIBUTES #
    ##############
    @property
    def database(self):
        '''
        Database to blast against.

        @returns: {String}
        '''
        return self._database

    @property
    def database_index(self):
        '''
        Index to correct database positions (if exists)

        @return: {String}
        '''
        return self._idx

    @property
    def self_hit(self):
        '''
        Capture self hit in the database.

        @returns: {Boolean}
        '''
        return self._self_hit

    @property
    def hitid_format(self):
        '''
        Amount of hit name to store

        @returns: {String}
        '''
        return self._hitid_format

    @property
    def overwrite(self):
        '''
        Overwrite old files.

        @returns: {Boolean}
        '''
        return self._overwrite

    @property
    def clean_files(self):
        '''
        Remove temporary files after process is finished.

        @returns: {Boolean}
        '''
        return self._clean_files

    ###########
    # METHODS #
    ###########
    @staticmethod
    def dynamic(executable, path, dbformater):
        '''
        Manually set the values for blast path, executable and db format.

        @param:    executable
        @pdef:     name of the executable file
        @ptype:    {String}

        @param:    path
        @pdef:     path to the executable file
        @ptype:    {String}

        @param:    dbformater
        @pdef:     name of the database format executable
        @ptype:    {String}
        '''
        BlastExe._set_dynamic_executable(executable, path)
        BlastExe._DBFORMATER = dbformater

    def add_attribute(self, attribute_value, attribute_id):
        '''
        Adds specific attributes to the blast execution.

        @param:    attribute_value
        @pdef:     value of the attribute to add.
        @ptype:    {String}. Transformed otherwise.

        @param:    attribute_id
        @pdef:     label of the attribute to add.
        @pclash:   labels 'db' and 'outfmt' are blocked, as they need to be
                   specified when creating the instance.
        @ptype:    {String}
        '''
        if attribute_id in BlastExe._BLOCKED:
            raise self._error.blocked_parameter(BlastExe._BLOCKED)

        self._parameters.append([str(attribute_value), attribute_id])

    def execute_query_seq(self, sequenceID = None, sequence          = None,
                          blast_input_file = None, blast_output_file = None,
                          work_directory   = os.getcwd()):
        '''
        Execute BLAST given a query sequence.

        @param:    sequenceID
        @pdef:     name of the query sequence.
        @pdefault: 'QuerySequence'
        @pclash:   If sequence is not provided, it assumes that the sequenceID
                   belongs to a protein in the database and, thus, it searches
                   for it. Either sequenceID or sequence needs to be provided.
        @ptype:    {String}

        @param:    sequence
        @pdef:     query sequence.
        @pdefault: _None_
        @pclash:   Either sequenceID or sequence needs to be provided.
        @ptype:    {String}

        @param:    blast_input_file
        @pdef:     name of the temporary fasta file to use as BLAST input.
        @pdefault: job.pid + clock + .tmp.fa
        @ptype:    {String}

        @param:    blast_output_file
        @pdef:     name of the temporary BLAST output file.
        @pdefault: job.pid + clock + .blast.xml.out
        @ptype:    {String}

        @param:    work_directory
        @pdef:     Directory to which the temporary files will be created.
        @pdefault: Current working directory.
        @ptype:    {String}

        @raises: {AttributeError} if neither sequenceID nor sequence are
                  provided or if sequenceID is a list of sequence names.
        @raises: {BlastError} in BLAST execution or output parsing errors.

        @returns: {BlastResult}
        '''
        if sequenceID is None and sequence is None:
            msg = 'Either a sequence or sequenceID is needed to perform the blast.'
            raise AttributeError(msg)

        if isinstance(sequenceID, (list, set, tuple)):
            msg = 'Blasts can only be executed one at a time due to XML output restrictions.'
            raise AttributeError(msg)

        sequenceID = 'QuerySequence' if sequenceID is None else sequenceID

        # Given only a code implies that the protein of interest is in the
        # database itself
        if sequence is None:
            grabbedSequence = self._database.retrieve(sequenceID)
            sequenceID      = grabbedSequence[0].id
            sequence        = grabbedSequence[0].sequence

        # All the sequence is unknown, it will crash blast
        if len(re.sub(r'[Xx]', '', sequence)) == 0:
            SBIg.warn(self, 'Created an empty BlastResult.')
            return BlastResult(query_name     = sequenceID,
                               query_sequence = sequence)

        Path.mkdir(work_directory)
        file_prefixes = ".".join([str(os.getpid()), str(int(time.clock()*100000))])
        file_prefixes = os.path.join(work_directory, file_prefixes)
        tmp_input     = file_prefixes + ".tmp.fa"
        tmp_output    = file_prefixes + ".blast.xml.out"

        tmp_input  = tmp_input  if blast_input_file  is None else os.path.join(work_directory, blast_input_file)
        tmp_output = tmp_output if blast_output_file is None else os.path.join(work_directory, blast_output_file)

        QueryFasta = Fasta.build(file_name = tmp_input, sequence_id = sequenceID,
                                 sequence  = sequence,  force       = True)

        self._execute(input_file = QueryFasta, output_file = tmp_output)

        blast_result = self._parse_blast(sequence, tmp_output)

        self._clean([tmp_input, tmp_output])

        return blast_result

    def execute_query(self, query_file, blast_output_file = None,
                      work_directory = os.getcwd()):
        '''
        Execute BLAST given a query sequence.

        @param:    query_file
        @pdef:     Fasta file with the query sequence.
        @pdefault: 'QuerySequence'
        @ptype:    {String} or {File} or {Fasta}

        @param:    blast_output_file
        @pdef:     name of the temporary BLAST output file.
        @pdefault: query_file.prefix + job.pid + .blast.xml.out
        @ptype:    {String}

        @param:    work_directory
        @pdef:     Directory to which the temporary files will be created.
        @pdefault: Current working directory.
        @ptype:    {String}

        @raises: {AttributeError} if query_file is multi-fasta.
        @raises: {BlastError} in BLAST execution or output parsing errors.

        @returns: {BlastResult}
        '''
        if isinstance(query_file, basestring) or isinstance(query_file, File):
            newFasta = Fasta(fasta_file = query_file)
        elif isinstance(query_file, Fasta):
            newFasta = query_file

        if newFasta.is_multifasta:
            msg = 'Blasts can only be executed one at a time due to XML output restrictions.'
            raise AttributeError(msg)

        # All the sequence is unknown, it will crash blast
        newFasta.load()
        query_sequence = newFasta.sequence
        if len(re.sub(r'[Xx]', '', query_sequence.sequence)) == 0:
            SBIg.warn(self, 'Created an empty BlastResult.')
            return BlastResult(query_name     = query_sequence.id,
                               query_sequence = query_sequence.sequence)

        Path.mkdir(work_directory)
        file_prefixes = ".".join([newFasta.file.prefix, str(os.getpid())])
        file_prefixes = os.path.join(work_directory, file_prefixes)
        tmp_output    = file_prefixes + ".blast.xml.out"

        tmp_output = tmp_output if blast_output_file is None else os.path.join(work_directory, blast_output_file)

        self._execute(input_file = newFasta, output_file = tmp_output)

        blast_result = self._parse_blast(newFasta.sequence.sequence, tmp_output)

        self._clean([tmp_output, ])

        return blast_result

    def clean_optional_parameters(self):
        '''
        Deletes all optional parameters added to the {BlastExe}.
        '''
        self._parameters = []

    ###################
    # PRIVATE METHODS #
    ###################
    def _execute(self, input_file, output_file):
        '''
        Executes BLAST.

        @param:    input_file
        @pdef:     file with the sequence to blast. (SINGLE FASTA)
        @ptype:    {Fasta}

        @param:    output_file
        @pdef:     name of the blast output file.
        @ptype:    {String}

        @raises: {BlastError}
        '''
        if not os.path.isfile(output_file) or self.overwrite:
            self._EXE.backup_command()

            # Adding fixed blast parameters
            self._EXE.add_attribute(self._database.file.full, '-db')
            self._EXE.add_attribute('5', '-outfmt')
            self._EXE.add_parameter('-lcase_masking')

            # Adding optional parameters
            for parameter in self._parameters:
                self._EXE.add_attribute(parameter[0], parameter[1])

            # Adding input-output
            self._EXE.add_attribute(input_file.file.full, '-query')
            self._EXE.add_attribute(output_file,          '-out')

            try:
                self._EXE.execute()
                self._EXE.restore_command()
            except SystemError as e:
                default_warning = bool(re.search(self._error.default_blast_warning(),  str(e)))
                seleno_warning  = bool(re.search(self._error.selenocysteine_warning(), str(e)))
                repeat_warning  = bool(re.search(self._error.repetitive_sequence_warning(), str(e)))
                if not default_warning and not seleno_warning and not repeat_warning:
                    raise self._error.blast_execution_failed(e)

    def _clean(self, files):
        '''
        If clean_files is _True_ and not in debug mode, it removes the
        temporary files.

        @param:    files
        @pdef:     list of files to remove
        @ptype:    {List}
        '''
        if self.clean_files and not SBIg.debug:
            for temp_file in files:
                os.unlink(temp_file)

    def _parse_blast(self, query_sequence, blast_output_file):
        '''
        Calls the parsing of the blast output.

        @param:    query_sequence
        @pdef:     sequence that has been searched against the database.
        @ptype:    {String}

        @param:    blast_output_file
        @pdef:     output file of blast.
        @ptype:    {String}
        '''
        return BlastParser.parse(query_sequence, blast_output_file,
                                 self._self_hit, self._hitid_format)

    def _check_database(self, database):
        '''
        Ensures that the given database to blast upon exists and that it is
        formated for blast.
        I also sets the index file for the database if it exists.

        @param:    database
        @pdef:     database to blast upon.
        @ptype:    {String}

        @returns: {Fasta} object pointed to the database fasta file.
        '''
        # Database file does not exist
        if not os.path.isfile(database):
            return self._error.database_does_not_exist(database)

        # Database is not formated (if dbformatexe is added in the
        # configuration path it will be auto-formated)
        formatdb_files = []
        formatdb_sufix = BlastExe._EXEC_TYPES[self._search_type]

        for sufix in formatdb_sufix:
            if not os.path.isfile(database + sufix):
                formatdb_files.append(database + sufix)

        if len(formatdb_files) > 0:
            try:
                self._format_database(database)
            except ConfigParser.NoOptionError as e:
                raise self._error.no_blast_format_exe(e)
            except SystemError as e:
                raise self._error.wrong_db_format(database, e)

        idx       = os.path.abspath(database) + ".idx"
        self._idx = idx if os.path.isfile(idx) else None

        return Fasta(fasta_file = database)

    def _format_database(self, database):
        '''
        Executes the blast script to format the database.

        @param:    database
        @pdef:     database to blast upon.
        @ptype:    {String}
        '''
        SBIg.warn(self, 'Formating {0} for blast.'.format(database))
        dbexe  = Executable(executable = BlastExe._DBFORMATER,
                            path       = self._EXE.path)

        dbexe.add_attribute(database,               '-in')
        dbexe.add_attribute(self._search_type[0:4], '-dbtype')

        SBIg.alert('debug', self, 'Executing command {0}\n'.format(dbexe))

        dbexe.execute()

    def _check_hitid_format(self, hitid):
        '''
        Ensures that the hitid_format requested is amongst the available.

        @param:    hitid
        @pdef:     format of the name of the hit. If given a wrong option,
                   it defaults to 'single'
        @pdefault: 'single'
        @poptions: 'single' -> first word of the name,
                   'double' -> first two words of the hit name,
                   'all'    -> all the text in the hit name
        @ptype:    {String}

        @returns: {String}
        '''
        return 'single' if hitid not in BlastExe._HIT_ID_FRM else hitid

    def _check_execution_type(self, search_type):
        '''
        Ensures that the blast is performed in a known search mode.

        @param:    search_type
        @pdef:     type of blast search
        @poptions: 'protein', 'nucleotide'
        @ptype:    {String}

        @return: {String}
        '''
        if search_type not in BlastExe._EXEC_TYPES:
            raise self._error.incorrect_search_type(search_type)
        return search_type


class BlastParser(object):
    '''
    Processes a blast xml formated output into a {BlastResult} object.
    '''
    __metaclass__ = ABCMeta

    @staticmethod
    def parse(query_sequence, blast_output_file, self_hit, hitid_format):
        '''
        Processes a blast xml formated output into a {BlastResult} object.

        @param:    query_sequence
        @pdef:     sequence of the query protein/nucleotide.
        @ptype:    {String}

        @param:    blast_output_file
        @pdef:     output file from BLAST.
        @ptype:    {String}

        @param:   self_hit
        @pdef:     when _True_ if the query is found in the database, it is
                   retrieved.
        @pdefault: _False_
        @ptype:    {Boolean}

        @param:    hitid_format
        @pdef:     format of the name of the hit. If given a wrong option,
                   it defaults to 'single'
        @pdefault: 'single'
        @poptions: 'single' -> first word of the name,
                   'double' -> first two words of the hit name,
                   'all'    -> all the text in the hit name
        @ptype:    {String}

        @raises: {BlastError} if there are problems while parsing the XML file.
        @returns: {BlastResult}
        '''
        f = File(blast_output_file)
        s = BeautifulSoup(f.read())

        h = BlastHeader(version    = str(s.find('blastoutput_version').string),
                        matrix     = str(s.find('parameters_matrix').string),
                        gap_open   = int(s.find('parameters_gap-open').string),
                        gap_extend = int(s.find('parameters_gap-extend').string),
                        database   = str(s.find('blastoutput_db').string),
                        self_hit   = self_hit)
        b = BlastResult(query_name     = str(s.find('blastoutput_query-def').string),
                        query_sequence = query_sequence,
                        header         = h)

        SBIg.alert('debug', BlastParser(), b.str_blast_header())

        error_bool = False
        error_str  = []
        for iteration in s.find_all('iteration'):
            iternum = int(iteration.find('iteration_iter-num').string)
            for hit in iteration.find_all('hit'):
                hit_name  = BlastParser.hit_name(str(hit.find('hit_def').string), hitid_format)
                hit_lenth = int(hit.find("hit_len").string)
                for subhit in hit.find_all("hsp"):
                    data = BlastParser.parse_subhit(subhit)
                    r = BlastHit(hit            = [hit_name, hit_lenth],
                                 sequences      = [data['qs'], data['hs'], data['sc']],
                                 sequence_inits = [data['qp'], data['hp']],
                                 iteration      = iternum,
                                 stats          = [data['hi'], data['h+'],
                                                   data['hg'], data['ev']])
                    if not BlastParser.same_query_hit_names(b.query, hit_name, self_hit):
                        dbug_info = 'Added hit {0} in iteration {1}'
                        SBIg.alert('debug', BlastParser(), dbug_info.format(hit_name, r.iteration))
                        b.add_hit(r)
                        if not r.are_segments_ok:
                            error_bool = True
                            error_str.append("Check the alignment's fragmentation")
                            error_str.append("for the query %s with %s\n".format(b.query, hit_name))
                            error_str.append("{0}\n".format(r))
        b.set_last_iteration()
        if error_bool:
            SBIg.warn(BlastParser(), error_str)
            be = BlastError()
            raise be.parse_error()
        return b

    @staticmethod
    def parse_subhit(subhit):
        '''
        Extracts the relevant data from a subhit representation.

        @param:    subhit
        @pdef:     XML section that represents a subhit
        @ptype:    {bs4.element.Tag}

        @returns: {Dictionary}
        '''
        data = {}

        gaps       = subhit.find("hsp_gaps")
        data['hg'] = int(gaps.string) if gaps is not None else 0
        idnt       = subhit.find("hsp_identity")
        data['hi'] = int(idnt.string) if idnt is not None else 0
        posi       = subhit.find("hsp_positive")
        data['h+'] = int(posi.string) if posi is not None else 0

        data['ev'] = float(subhit.find("hsp_evalue").string)
        data['al'] = int(subhit.find("hsp_align-len").string)
        data['qs'] = str(subhit.find("hsp_qseq").string).strip()
        data['hs'] = str(subhit.find("hsp_hseq").string).strip()
        data['qp'] = int(subhit.find("hsp_query-from").string)
        data['hp'] = int(subhit.find("hsp_hit-from").string)
        data['sc'] = str(subhit.find("hsp_midline").string).strip()

        return data

    @staticmethod
    def same_query_hit_names(query, hit, self_hit):
        '''
        Compares the query and hit names to check if they are the same.
        Comparison is based on the first section of the name.
        If self_hit is _True_ it always assumes that they are NOT the same.

        @param:    query
        @pdef:     query name.
        @ptype:    {String}

        @param:    hit
        @pdef:     hit name.
        @ptype:    {String}

        @param:    self_hit
        @pdef:     tag to decide whether to compare the names.
        @ptype:    {Boolean}
        '''
        if self_hit:
            return False
        return query.split()[0].strip() == hit.split()[0].strip()

    @staticmethod
    def hit_name(full_name, hitid_format):
        '''
        Processes the name of the hit depending on the selected format.

        @param:    full_name
        @pdef:     original name of the hit
        @ptype:    {String}

        @param:    hitid_format
        @pdef:     format expected for the hit name
        @ptype:    {String}

        '''
        if hitid_format   == 'single':
            return full_name.split()[0].strip()
        elif hitid_format == 'double':
            return " ".join(full_name.split()[:2]).strip()
        else:
            return full_name


class BlastError(Exception):
    '''
    Manages different errors that can occur during the blast execution or
    parsing.

    '''
    _MSG     = ''
    _EXE     = False
    _NOBLAST = 'Blast has NOT been executed.'

    def __init__(self):
        pass

    def incorrect_search_type(self, qsearch):
        BlastError._EXE = False
        BlastError._MSG = '{0} is not a valid blast search method.'.format(qsearch)
        return self

    def database_does_not_exist(self, database):
        BlastError._EXE = False
        BlastError._MSG = '{0} is not a file.'.format(database)
        return self

    def wrong_db_format(self, database, e):
        BlastError._EXE = False
        BlastError._MSG = '{0} could not be formated for blast.'.format(database)
        BlastError._MSG += '\n' + str(e)
        return self

    def no_blast_format_exe(self, e):
        BlastError._EXE = False
        BlastError._MSG = 'Executable to format blast is not specified.'
        BlastError._MSG += '\n' + str(e)
        return self

    def blast_execution_failed(self, e):
        BlastError._EXE = False
        BlastError._MSG = 'Blast execution failed.'
        BlastError._MSG += '\n' + str(e)
        return self

    def blocked_parameter(self, param):
        BlastError._EXE = False
        BlastError._MSG = '{0} are blocked.'.format(param)
        return self

    def parse_error(self):
        BlastError._EXE = True
        BlastError._MSG = 'An error occurred while parsing the BLAST output.'
        return self

    def corrected_hits(self):
        BlastError._EXE = True
        BlastError._MSG = 'Hits have already been corrected.'
        return self

    def default_blast_warning(self):
        msg  = 'Warning: Composition-based score adjustment conditioned on '
        msg += 'sequence properties and unconditional composition-based score '
        msg += 'adjustment is not supported with PSSMs, resetting to default '
        msg += 'value of standard composition-based statistics'
        return msg

    def selenocysteine_warning(self):
        return 'Selenocysteine \(U\) at position'

    def repetitive_sequence_warning(self):
        return 'bad probabilities from'

    def __str__(self):
        d  = BlastError._MSG + '\n'
        d += BlastError._NOBLAST if not BlastError._EXE else ''
        return d



import sys
import argparse
import os
import re
import ConfigParser
import shutil
import subprocess
import time
import random
from collections import Counter

# Add '.' to sys.path
src_path = os.path.abspath(os.path.dirname(__file__))
sys.path.append(src_path)

# Read configuration file
config = ConfigParser.ConfigParser()
config_file = os.path.join(src_path, 'config.ini')
config.read(config_file)

# Add SBI library path to sys.path
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','sbi_library_path'))
sys.path.append(src_path)

# Add functions path to sys.path
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','functions_path'))
sys.path.append(src_path)

# Import functions
from SeqIO import *
from functions import *

def main():
    
    #initialize
    options    = parse_user_arguments()
    verbose    = options.show
    dummy_dir  = options.dummy_dir
    if not os.path.exists(options.dummy_dir): os.makedirs(options.dummy_dir)
    pdb_file   = options.pdb_file
    fasta_file = options.fasta_file
    chains_file= options.chains_file
    hydrogens  = options.hbplus
    outfile    = options.outfile

    if not fileExist(pdb_file) or not fileExist(fasta_file) or not fileExist(chains_file):
       sys.stderr.write('EXIT: Missing files, please check you have the files of PDB, FASTA and list of chains\n')
       exit(0)

    path_pdb = "/".join(os.path.abspath(pdb_file).split("/")[0:-1])
    pdb_name = pdb_file.split("/")[-1]

    proteins= {}
    if verbose: sys.stdout.write('\t\t\t-- Reading protein sequences...\n')
    for protein in FASTA_iterator(fasta_file):
            name = protein.get_identifier()
            if len(name.split("|"))>2:
               name1 = name.split("|")[1] 
               name2 = name.split("|")[2] 
            if len(name.split("|")) == 2:
               name1 = name.split("|")[1]
               name2 = name.split("|")[1]
            if len(name.split("|")) == 1:
               name1 = name.split("|")[0]
               name2 = name.split("|")[0]
            seq  = protein.get_sequence()
            if len(seq) > 0:
                try:
                    proteins.setdefault(name1, ProteinSequence(name1, seq))
                    proteins.setdefault(name2, ProteinSequence(name2, seq))
                except IncorrectSequenceLetter as e:
                    sys.stderr.write('WARNING: %s\n' %e)
                    sys.stdout.write('\t\t\t-- Skip input sequence: %s Sequence: %s\n' %(name, seq))

    sequences={}
    fd=open(chains_file,"r")
    for line in fd:
       chain,fasta_name = line.strip().split()
       if proteins.has_key(fasta_name):
          sequences.setdefault(chain,proteins[fasta_name])
          #print "%s\t%s\t%s\t%s\n"%(chain,fasta_name,proteins[fasta_name].get_identifier(),proteins[fasta_name].get_sequence())
       else:
          sys.stderr.write('EXIT: Missing FASTA sequence %s\n'%fasta_name)
          exit(0)
    fd.close()
   
    if hydrogens:
       pdb_hydro=".".join(pdb_name.split(".")[0:-1])+".h"
       if not fileExist(os.path.join(path_pdb,pdb_hydro)):
            try:
               if verbose: sys.stdout.write("\t\t\t-- Adding hydrogens and relaxing the model %s\n"%pdb_name)
               add_hydrogens(config,path_pdb,pdb_name, pdb_hydro)
            except ValueError as e:
               sys.stderr.write("WARNING %s\n"%e)
               shutil.copy(pdb_file, os.path.join(path_pdb,pdb_hydro))
       else:
            if verbose: sys.stdout.write("\t\t\t-- The model has already hydrogens %s\n"%pdb_hydro)
       input_renumber=pdb_hydro
    else:
       input_renumber=pdb_name

    if not fileExist(outfile):
     pdb_new=PDB()
     pdb_new=renumber_pdb(config,path_pdb,input_renumber,sequences,dummy_dir)
     pdb_new.write(outfile)
    else:
     if verbose: sys.stdout.write("\t\t\t-- File %s already exist\n"%outfile)
        
def parse_user_arguments(*args, **kwds):
    parser = argparse.ArgumentParser(
        description = 'Automatic Renumbering of a protein with the original sequence(s)',
        epilog      = '@Oliva\'s lab 2016')
    parser.add_argument('-l', '--list', dest = 'chains_file', action = 'store',
                        help = 'Input file with a list of pairs "CHAIN FASTA_ID" to associate the chain name in the PDB file and the name of the sequence')
    parser.add_argument('-f', '--fasta', dest = 'fasta_file', action = 'store',
                        help = 'Input file with a list of FASTA sequences of the proteins to use')
    parser.add_argument('-p', '--pdb', dest = 'pdb_file', action = 'store',
                        help = 'Input file with a list of FASTA sequences of the proteins to use')
    parser.add_argument('-o', '--out', dest = 'outfile', action = 'store',
                        help = 'PDB output file renumbered')
    parser.add_argument('-d', '--dummy_dir', dest = 'dummy_dir', action = 'store', default = '/tmp/modppi_dummy',
                        help = 'Specifies the location of the dummy folder (default is /tmp/modppi_dummy)')
    parser.add_argument('-hydro','--hydrogens', dest = 'hbplus', action = 'store_true',
                        help = 'Flag to include hydrogens (default is False)')
    parser.add_argument('-v', '--verbose', dest = 'show', action = 'store_true',
                        help = 'Flag for verbose mode (default is False)')
    options = parser.parse_args()
    return options
    
# MAIN ###########################################################################################################

if __name__ == '__main__':
    main()



"""
AminoAtom

author: jbonet
date:   02/2013

@oliva's lab
"""

from . import Atom

class AtomOfAminoAcid(Atom):
    """
    An {AtomOfAminoAcid} is simply a point in space defined by 3 coordinates
    WITH specific functions for atoms in amino acids
    """

    backbone_atoms = set(['N', 'CA', 'C'])

    #
    # BOOLEANS
    #
    @property
    def is_Calpha(self): return self._name == "CA"

    @property
    def is_Cbeta(self):  return self._name == "CB"

    @property
    def is_N(self):      return self._name == "N"

    @property
    def is_C(self):      return self._name == "C"

    @property
    def is_O(self):      return self._name == "O"




class SeqAliError(Exception):
    """
    Complementary Error Class: SeqAliError
    """
    def __init__(self, code = None, value = None):

        self.value = value
        self.code  = code

        if self.get_code() is not None:
            self._determine_message()

    def get_value(self):        return self.value
    def get_code(self):         return self.code

    def _determine_message(self):

        if self.get_code() ==  1:
            self.value = "Something has gone wrong with the sequence fragment detection\n"

        elif self.get_code() ==  2:
            self.value = "A multiple sequence alignment of {0} sequences cannot be evaluated with Rost's Twilight Zone\n".format(self.value)

    def __str__(self):
        error_str = "\n\n ###################\n[SeqAliError Type: %d]\n ###################\n%s\n" %(self.get_code(), self.get_value())
        return error_str


'''
@file: Path.py

@author: Jaume Bonet
@mail:   jaume.bonet@gmail.com
@date:   02/2013

@ [oliva's lab](http://sbi.imim.es)

@class: Path
'''
import os
import fnmatch
import shutil

from .  import FileError
from .. import SBIglobals as SBIg


class Path(object):
    '''
    Collection of static methods (can be called without declaring an instance
    of the class) to help in file and directory management.
    They are defined this way and not as single methods to avoid overwrite
    system functions.
    '''

    @staticmethod
    def list_files(root              = os.getcwd(), pattern  = '*',
                   avoid_empty_files = True,        rootless = False):
        '''
        Returns all file names in a directory tree matching a specific pattern.

        @param:    root
        @pdef:     Root of the directory tree to search.
        @pdefault: Current working directory.
        @ptype:    {String}

        @param:    pattern
        @pdef:     Expression(s) to match (ls-like format)
        @pdefault: '*'
        @ptype:    {String} or {List}

        @param:    avoid_empty_files
        @pdef:     Ignore files with size 0.
        @pdefault: _True_
        @ptype:    {Boolean}

        @param:    rootless
        @pdef:     When False, the name of the files are returned with
                   absolute path. Otherwise the root is removed.
        @pdefault: _False_
        @ptype:    {Boolean}

        @raises: {FileError} if root is a file
        @yields: {String}.
        '''

        if os.path.isfile(root):
            error = FileError(root)
            raise error.is_file()

        search_patterns = []
        if not isinstance(pattern, list):
            search_patterns.append(pattern)
        else:
            search_patterns = pattern

        for pat in search_patterns:
            for path, dirs, files in os.walk(os.path.abspath(root)):
                for filename in fnmatch.filter(files, pat):
                    qfile = os.path.join(path, filename)
                    size = os.path.getsize(qfile)
                    if not avoid_empty_files or size > 0:
                        msg = 'Found file {0}'.format(qfile)
                        SBIg.alert('debug', Path(), msg)
                        if not rootless:
                            yield qfile
                        else:
                            root = os.path.abspath(root) + "/"
                            yield qfile.replace(root, '')

    @staticmethod
    def list_directories(root = os.getcwd(), rootless = False):
        '''
        Returns all directory names in a directory tree.

        @param:    root
        @pdef:     Root of the directory tree to search.
        @pdefault: Current working directory.
        @ptype:    {String}

        @param:    rootless
        @pdef:     When False, the name of the directories are returned with
                   absolute path. Otherwise the root is removed.
        @pdefault: _False_
        @ptype:    {Boolean}


        @raises: {FileError} if root is a file
        @yields: {String}.
        '''

        if os.path.isfile(root):
            error = FileError(root)
            raise error.is_file()

        for path, dirs, files in os.walk(os.path.abspath(root)):
            for onedir in dirs:
                qdir = os.path.join(path, onedir)
                SBIg.alert('debug', Path(), 'Found directory {0}'.format(qdir))
                if not rootless:
                    yield qdir
                else:
                    r = root if root.endswith('/') else root + '/'
                    yield qdir.replace(r, '')

    @staticmethod
    def sync_directories(source_dir, destination_dir, by_dir = True,
                         by_file = False, only_list = False):
        '''
        Synchronizes two directory trees.

        @param:    source_dir
        @pdef:     Name of the original directory.
        @ptype:    {String}

        @param:    destination_dir
        @pdef:     Name of the synchronized directory.
        @ptype:    {String}

        @param:    by_dir
        @pdef:     Sync is performed at directory level.
        @pdefault: _True_
        @pclash:   by_file
        @ptype:    {Boolean}

        @param:    by_file
        @pdef:     Sync is performed at file level.
        @pdefault: _False_
        @pclash:   by_dir
        @ptype:    {Boolean}

        @param:    only_list
        @pdef:     When True, sync is not performed, only listed.
        @pdefault: _False_
        @ptype:    {Boolean}

        @raises: {AttributeError} if by_dir and by_file are both True.
        @raises: {FileError} if either source_dir or destination_dir are files.
        @yields: {String}. Names of the synchronized files/directories
        '''

        if by_file is True and by_dir is True:
            raise AttributeError('Both sync methods can not be active\n')
        if os.path.isfile(source_dir):
            error = FileError(source_dir)
            raise error.is_file()
        if os.path.isfile(destination_dir):
            error = FileError(destination_dir)
            raise error.is_file()

        sdir = os.path.abspath(source_dir)
        ddir = os.path.abspath(destination_dir)
        pdir = '#not_a_dir_at_all'

        source_dirs = set()
        if by_dir:
            for onedir in Path.list_directories(root = sdir, rootless = True):
                source_dirs.add(onedir)
            for onedir in Path.list_directories(root = ddir, rootless = True):
                if onedir in source_dirs:
                    source_dirs.remove(onedir)
        if by_file:
            for onedir in Path.list_files(root = sdir, rootless = True):
                source_dirs.add(onedir)
            for onedir in Path.list_files(root = ddir, rootless = True):
                if onedir in source_dirs:
                    source_dirs.remove(onedir)

        for onedir in sorted(source_dirs):
            onedir = onedir.lstrip('/')
            if not os.path.commonprefix([pdir, onedir]) == pdir:
                msg = '{0} is different from {1} to {2}'.format(onedir,
                                                                sdir, ddir)
                SBIg.alert('verbose', Path(), msg)
                fullori = os.path.join(sdir, onedir)
                fullnew = os.path.join(ddir, onedir)
                if not only_list:
                    shutil.copytree(fullori, fullnew)
                    yield fullori
                else:
                    yield fullori
                pdir = onedir

    @staticmethod
    def do_files_match(root = os.getcwd(), pattern = '*',
                       avoid_empty_files = True):
        '''
        Determines if there is any file in a directory tree matching a
        specific pattern.

        @param:    root
        @pdef:     Root of the directory tree to search.
        @pdefault: Current working directory.
        @ptype:    {String}

        @param:    pattern
        @pdef:     Expression(s) to match (ls-like format)
        @pdefault: '*'
        @ptype:    {String} or {List}

        @param:    avoid_empty_files
        @pdef:     Ignore files with size 0.
        @pdefault: _True_
        @ptype:    {Boolean}

        @raises: {FileError} if root is a file
        @return: {Boolean}
        '''
        if os.path.isfile(root):
            error = FileError(root)
            raise error.is_file()

        search_patterns = []
        if not isinstance(pattern, list):
            search_patterns.append(pattern)
        else:
            search_patterns = pattern

        for pat in search_patterns:
            for path, dirs, files in os.walk(os.path.abspath(root)):
                for filename in fnmatch.filter(files, pat):
                    size = os.path.getsize(os.path.join(path, filename))
                    if not avoid_empty_files or size > 0:
                        return True
        return False

    @staticmethod
    def mkdir(new_dir):
        '''
        Creates a new directory.
        Ignores the command if the directory already exists.
        Creates parent directories as required.

        @recipe: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/82465

        @param:    new_dir
        @pdef:     Name of the new directory.
        @ptype:    {String}

        @raises: {FileError} if the directory can not be created
        '''
        error = FileError(os.path.abspath(new_dir))
        if os.path.isdir(new_dir):
            SBIg.warn(Path(), 'A directory with the name {0} already exists.'.format(new_dir))
        elif os.path.isfile(new_dir):
            raise error.is_file()
        else:
            SBIg.alert('debug', Path(), 'Creating dir {0}.'.format(new_dir))
            head, tail = os.path.split(new_dir)
            if head and not os.path.isdir(head):
                Path.mkdir(head)
            if tail:
                os.mkdir(new_dir)

    @staticmethod
    def copy_file(source, destination, overwrite = False):
        '''
        Copy a file from one place to another.

        @param:    source
        @pdef:     File to copy.
        @ptype:    {String}

        @param:    destination
        @pdef:     Copy file.
        @ptype:    {String}

        @param:    overwrite
        @pdef:     Overwrite files if they already exist.
        @pdefault: _SBIglobals.overwrite_
        @ptype:    {Boolean}

        @raises: {FileError} if source does not exist or destination exists
                 and can not be overwritten
        '''
        # Local overwrite takes precedence over Global overwrite
        overwrite = SBIg.decide_overwrite(overwrite)

        if not os.path.exists(source):
            error = FileError(source)
            raise error.file_does_not_exist()
        if os.path.exists(destination) and not overwrite:
            error = FileError(destination)
            raise error.file_overwrite()

        shutil.copyfile(src = source, dst = destination)



import numpy as np
from SBI.beans.JSONer import JSONer


class BioMolecule(JSONer):
    def __init__(self, identifier):
        self._identifier = identifier
        self._matrices   = []
        self._chains     = set()

    #
    # ATTRIBUTES
    #
    @property
    def identifier(self):
        return self._identifier

    @property
    def chains(self):
        return self._chains

    @chains.setter
    def chains(self, value):
        if isinstance(value, list):
            self._chains.update(set(value))
        else:
            self._chains.add(value)

    @property
    def matrices(self):
        if len(self._matrices) > 0:
            return self._matrices
        else:
            return [MatrixAndVector.stillmatrix(), ]

    #
    # FUNCTIONS
    #
    def new_matrix(self):
        self._matrices.append(MatrixAndVector.stillmatrix())

    def update_last_matrix(self, row, mx, my, mz, v):
        self._matrices[-1].update_matrix(row, mx, my, mz)
        self._matrices[-1].update_vector(row, v)

    #
    # OUTPUT FUNCTIONS
    #
    def as_dict(self):
        data = {'id':       self.identifier,
                'chains':   list(self.chains),
                'matrices': [x.as_dict() for x in self.matrices]}

        return data

    #
    # OVERWRITE DEFAULT FUNCTIONS
    #
    def __len__(self):
        return len(self._matrices)


class MatrixAndVector(object):
    def __init__(self, matrix, vector):
        self._matrix = matrix
        self._vector = vector

    #
    # ATTRIBUTES
    #
    @property
    def matrix(self):
        return self._matrix

    @property
    def vector(self):
        return self._vector

    #
    # FUNCTIONS
    #
    @staticmethod
    def stillmatrix():
        return MatrixAndVector(np.identity(3, float),
                               np.zeros(3, float))

    def update_matrix(self, row, mx, my, mz):
        self._matrix[int(row)-1][0] = float(mx)
        self._matrix[int(row)-1][1] = float(my)
        self._matrix[int(row)-1][2] = float(mz)

    def update_vector(self, row, v):
        self._vector[int(row)-1] = float(v)

    #
    # OUTPUT FUNCTIONS
    #
    def as_dict(self):
        data = {'matrix': self.matrix.tolist(),
                'vector': self.vector.tolist()}
        return data



from SBI.external.CDhit import CDhitList

tables = {'chain' : 'chain',
          'homo'  : 'chain_homology'}

class CDhit(CDhitList):
    def toSQL(self):
        command = ''
        for hit in self.clusters:
            mpdb, mchain = hit.master.name.split('_')[0], hit.master.name.split('_')[1]
            command += "SET @master = (SELECT nid FROM {0} WHERE pdb='{1}' AND chain='{2}');\n".format(tables['chain'], mpdb,mchain)
            command += "UPDATE {0} SET clustered = 'M' WHERE nid=@master;\n".format(tables['chain'])
            for s in hit.sequences:
                seq = hit.sequences[s]
                spdb, schain = seq.name.split('_')[0], seq.name.split('_')[1]
                command += "SET @homo = (SELECT nid FROM {0} WHERE pdb='{1}' AND chain='{2}');\n".format(tables['chain'], spdb,schain)
                command += "UPDATE {0} SET clustered = 'H' WHERE nid=@homo;\n".format(tables['chain'])
                command += "INSERT INTO {0} VALUES (@master, @homo, {1.length}, {2.length}, {2.homology});\n".format(tables['homo'], hit.master, seq)
        return command



class CDhitHomolog(object):
    '''
    Representative of a sequence in the cd-hit cluster.

    '''
    def __init__(self, name, length, homology):
        '''
        @param:    name
        @pdef:     name of the sequence
        @ptype:    {String}

        @param:    length
        @pdef:     length of the sequence
        @ptype:    {String}

        @param:    homology
        @pdef:     level of homology inside the cluster
        @ptype:    {String}
        '''
        self._name     = name.lstrip('>').rstrip('.')
        self._length   = int(length.replace('aa', '').rstrip(','))
        self._homology = homology.replace('%', '')

    ##############
    # ATTRIBUTES #
    ##############
    @property
    def name(self):
        '''
        name of the sequence

        @return: {String}
        '''
        return self._name

    @property
    def length(self):
        '''
        length of the protein.

        @return: {Integer}
        '''
        return self._length

    @property
    def homology(self):
        '''
        level of homology inside the cluster

        @return: {Integer} or {String}
        '''
        if self.is_master:
            return self._homology
        else:
            return int(self._homology)

    @homology.setter
    def homology(self, value):
        '''
        @param:    value
        @pdef:     level of homology inside the cluster
        @ptype:    {String}
        '''
        self._homology = int(value)

    ############
    # BOOLEANS #
    ############
    @property
    def is_master(self):
        '''
        checks if protein is representative of the cluster.

        @return: {Boolean}
        '''
        return self._homology == '*'

    def __repr__(self):
        if not self.is_master:
            return '{0.name}: {0.length:0004d} Aa with {0.homology:003d}%'.format(self)
        else:
            return '{0.name}: {0.length:0004d} Aa.'.format(self)



"""
ContactAN

author: jbonet
date:   02/2014

@oliva's lab
"""
from .   import Contact
from SBI import SBIglobals

class ContactAN(Contact):

    available_distance_types = set(["min","geometric","cbbackbone"])
    atomic_types             = set(["min","cbbackbone"])
    description              = 'AminoNucleo'

    def __init__(self, aminoacid, nucleotide, threshold_type = "min", threshold_distance = 8):

        self._check_type(requested_type = threshold_type)

        self._threshold_type     = threshold_type
        self._threshold_distance = threshold_distance

        super(ContactAN, self).__init__(residue1 = aminoacid, residue2 = nucleotide)

    #
    # ATTRIBUTES
    #
    @property
    def aminoacid(self):            return self._residue1

    @property
    def nucleotide(self):           return self._residue2

    @property
    def min_distance(self):         return self._distance["min"][2]
    @property
    def min_atoms(self):            return self._distance["min"][0:2]

    @property
    def geometric_distance(self):   return self._distance["geometric"][2]

    @property
    def cbbackbone_distance(self):  return self._distance["cbbackbone"][2]
    @property
    def cbbackbone_atoms(self):     return self._distance["cbbackbone"][0:2]

    #
    # OVERWRITE PARENT FUNCTION
    #
    def _build(self):
        for dist_type in self.available_distance_types:
            self._distance.setdefault(dist_type, None)

        self._distance[self._threshold_type] = self.aminoacid.distance(self.nucleotide, dist_type = self._threshold_type)
        SBIglobals.alert('deepdebug', self, '\tEvaluating distance {0:.3f} of {1}'.format(self._distance[self._threshold_type][2], self._threshold_type))
        if float(self._distance[self._threshold_type][2]) <= self._threshold_distance and \
           float(self._distance[self._threshold_type][2]) >= 0:
            SBIglobals.alert('deepdebug', self, '\tDistance under threshold.')
            self._underthreshold = True
            for dist_type in self._distance:
                if dist_type != self._threshold_type:
                    SBIglobals.alert('deepdebug', self, '\tGathering {0} distance'.format(dist_type))
                    self._distance[dist_type] = self.aminoacid.distance(self.nucleotide, dist_type = dist_type)
                    SBIglobals.alert('deepdebug', self, '\t\tDistance {0:.3f} of {1}'.format(self._distance[dist_type][2], dist_type))
            

    #
    # TOSTRING
    #
    def toString(self, requested_type = "cb"):
        if requested_type != "all":
            self._check_type(requested_type = requested_type)
            ddata = self._distance[requested_type]
            if requested_type in self.atomic_types:
                return '{0.aminoacid.type}:{0.aminoacid.number}:{1[0].name}\t{0.nucleotide.type}:{0.nucleotide.number}:{1[1].name}\t{1[2]:.3f}'.format(self, ddata)
            else:
                return '{0.aminoacid.type}:{0.aminoacid.number}\t{0.nucleotide.type}:{0.nucleotide.number}\t{1[2]:.3f}'.format(self, ddata)
        else:
            return '{0.aminoacid.type}:{0.aminoacid.number}\t{0.nucleotide.type}:{0.nucleotide.number}\t{0.min_distance:.3f}\t{0.geometric_distance:.3f}\t{0.cbbackbone_distance:.3f}'.format(self)




"""StorableObject

author: jbonet
date:   01/2013

@oliva's lab
"""

"""
Import Standard Libraries
"""
from abc import ABCMeta
try:     import cPickle as pickle
except:  import pickle
import sys

"""
Dependences in SBI library
"""
from .   import File
from SBI import SBIglobals

class StorableObject(object):
    """StorableObject is an abstract "dumping" class.

        This means that it is basically usefull for those who would like to extend this library.
        Basically, it gives the object the hability to be "dumped" on disk and be recovered afterwards.

        [AS AN ABSTRACT OBJECT, IT DOES NOT HAVE A CONSTRUCTOR]

        Methods:
            - dump(): Stores the object into a file
                        - object_file (string): Name for the output file
                                                @Mandatory
                        - overwrite (bool):     Overwrite previous file of the same name
                      @Raises FileError

            > load(): Retrieves the object from a python object file
                        - object_file (string): Name of the file containing the object
                                                @Mandatory
                      @Returns the loaded object.
                      @staticmethod: can be called without any instance declared
                      @Raises FileError
                      
        Requirements:
            * abc, cPickle/pickle, sys
            < SBI.beans.File
            < SBI.SBIglobals
    """
    __metaclass__ = ABCMeta
    
    def dump(self, object_file, overwrite = None):
        """
            - dump(): Stores the object into a file
                        - object_file (string): Name for the output file
                                                @Mandatory
                        - overwrite (bool):     Overwrite previous file of the same name
                      @Raises FileError
        """
        SBIglobals.alert('verbose', self, 'Writting object to file {0}'.format(object_file))
        dumpFile = File(file_name = object_file, action = 'wb', overwrite = overwrite)
        pickle.dump(self, dumpFile.descriptor)
        dumpFile.close()
    
    @staticmethod
    def load(object_file):
        """
            > load(): Retrieves the object from a python object file
                        - object_file (string): Name of the file containing the object
                                                @Mandatory
                      @Returns the loaded object.
                      @staticmethod: can be called without any instance declared
                      @Raises FileError
        """
        SBIglobals.alert('verbose', StorableObject, 'Preparing to load object from file {0}'.format(object_file))
        Object   = None
        loadFile = File(file_name = object_file, action='rb')
        Object   = pickle.load(loadFile.descriptor) 
        loadFile.close()
        return Object




import mysql.connector

class ArchDB(object):
    def __init__(self, dbname=None, dbhost=None, dbuser=None, dbpassword=None):
        self._database  = dbname
        self._host      = dbhost
        self._user      = dbuser
        self._password  = dbpassword

        self._db        = None
        self._cursor    = None

        self._connect()


    #
    # PRIVATE FUNCTIONS
    #

    def _connect(self):

        # Connection to MySQL server 
        if not self.dbuser is None and not self.dbpassword is None:
            self._db = mysql.connector.connect(host= self.dbhost, user=self.dbuser, passwd=self.dbpassword)
        elif not self.dbuser is None:
            self._db = mysql.connector.connect(host= self.dbhost, user=self.dbuser)
        else:
            self._db = mysql.connector.connect(host= self.dbhost)

        # Cursor (from where operations are done and information is retrieved)
        self._cursor = self.db.cursor()

        # Specify Database 
        if( self.dbname is not None ):
            self.db.database=self.dbname


import re, copy
import BioLib.Structure.PDB as PDB

class PATCHDock(object):
    '''
    Class to manage the PACHDock transform file
    '''
    def __init__(self, PACHDock_file):
        '''
        Call the parsing method
        '''
        self.decoys = []
        self.__parse(PACHDock_file)

    def __parse(self, PACHDock_file):
        '''
        '''
        PACHDock_fo = open(PACHDock_file, 'r')
        for decoy in PACHDock_fo:
            if '#' not in decoy and '||' in decoy:
                decoy = decoy.strip('\n').split('||')
                num = int(decoy[0].split('|')[0])
                score = int(decoy[0].split('|')[1])
                pen = float(decoy[0].split('|')[2])
                area = float(decoy[0].split('|')[3])
                as12 = int(decoy[0].split('|')[6])
                ace = float(decoy[0].split('|')[7])
                vectors = re.split('\s+', decoy[1])
                self.decoys.append(PATCHDecoy(num, score, pen, area, as12, ace,
                                              float(vectors[1]), float(vectors[2]), float(vectors[3]), 
                                              float(vectors[4]), float(vectors[5]), float(vectors[6])))
        PACHDock_fo.close()

    # GETTERS #

    def get_decoys(self):
        '''
        Return all decoys
        '''
        return self.decoys

    def get_decoy(self, num):
        '''
        Return a decoy by its number
        '''
        for decoy in self:
            if decoy.get_num() == num:
                return decoy
        return None

    # METHODS #

    def get_num_decoys(self):
        return len(self.decoys)

    def print_structures(self, static_structure, mobile_structure, pdb_file, single_pdb=False):
        '''
        Print all PDB decoys
        '''
        decoy_list = []
        for decoy in self:
            if singlePDB:
                decoy_structure = decoy.get_structure(mobile_structure)
                decoy_list.append((static_structure, decoy_structure))
            else:
                decoy.print_structure(static_structure, decoy_structure, pdb_file+'_'+str(decoy.get_num())+'.pdb')
        if singlePDB:
            PDB.write_pdb(decoy_list, pdb_file+'.pdb', multi_chain=True, multi_model=True)

    # MAGIC METHODS #

    def __iter__(self):
        '''
        Iterates over each decoy
        '''
        for decoy in self.decoys:
            yield decoy

    def __str__(self):
        '''
        toString method
        '''
        return '\n'.join([decoy.__str__() for decoy in self.decoys])

class PATCHDecoy(object):
    '''
    Represents a decoy result from a PACHDock transform file
    '''
    def __init__(self, num, score, pen, area, as12, ace, alpha, beta, gamma, x, y, z):
        '''
        '''
        self.num = num
        self.score = score
        self.pen = pen
        self.area = area
        self.as12 = as12
        self.ace = ace
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.x = x
        self.y = y
        self.z = z

    # GETTERS #

    def get_num(self):
        return self.num

    def get_score(self):
        return self.score

    def get_pen(self):
        return self.pen

    def get_area(self):
        return self.area

    def get_as12(self):
        return self.as12

    def get_ace(self):
        return self.ace

    def get_tranlation_vector(self):
        return (self.x, self.y, self.z)
   
    def get_rotation_vector(self):
        return (self.alpha, self.beta, self.gamma)

    def get_structure(self, mobile_structure):
        '''
        Gets the structure of the given docking result
        '''
        structure = copy.deepcopy(mobile_structure)
        structure.relocate(self.get_tranlation_vector(), self.get_rotation_vector(), docking='PATCHDOCK') 
        return structure

    def print_structure(self, static_structure, mobile_structure, pdb_file):
        decoy_structure = self.get_structure(mobile_structure)
        PDB.write_pdb((static_structure, decoy_structure), pdb_file+'.pdb', multi_chain=True, multi_model=False)

    def __str__(self):
        return '%d %d %f %f %f %f %f %f' % (self.num, self.score, self.alpha, self.beta, self.gamma, self.x, self.y, self.z)



"""FileError

author: jbonet
date:   09/2013

@oliva's lab
"""
class FileError(Exception):
	"""Manages different errors produced by File when trying to Read/Write"""

	def __init__(self, code, choice = None, options = None):
		"""Init. Defines kinds of error"""
		self._code    = code
		self._choice  = choice
		self._options = options

	def _code_to_message(self):
		"""Pick and choose message"""
		if self._code == 0:	# One attribute is missing
			return "File object initialization error. A file_name must be given."
		if self._code == 1:	# Unknown action requested
			return "The requested action {0} is not implemented.\n".format(self._choice) + \
				   "The implemented actions are {0}.\n".format(self._options)
		if self._code == 2:	# Closing something not open
			return "Trying to close a filehandle that was not open for file {0}\n".format(self._choice)
		if self._code == 3:	# File existence exist
			if self._options == 'noexists':
				return "The file {0} does not exist.\n".format(self._choice)
			if self._options == 'exists':
				return "The file {0} already exists.\n".format(self._choice)
			if self._options == 'dirnoexists':
				return "The dir {0} does not exist.\n".format(self._choice)
			if self._options == 'filewithname':
				return "A file {0} already exists. Can not create directory.\n".format(self._choice)
		if self._code == 4:	# Permissions
			if self._options == 'read':
				return "Unable to read {0}. No permissions.\n".format(self._choice)
			if self._options == 'write':
				return "Unable to write in {0}. No permissions.\n".format(self._choice)
			if self._options == 'nodir':
				return "Unable to write in {0}. Directory does not exist.\n".format(self._choice)
		if self._code == 5:
			return "Attribute Error on object {0}, function {1}\n".format(self._choice, self._options)
		if self._code == 6: # Wrong action: write
			return "Unable to write. File not in 'write' mode.\n"

	def __str__(self):
		"""Final error reply"""
		return self._code_to_message()



'''
@file: StorableObject.py

@author: Jaume Bonet
@mail:   jaume.bonet@gmail.com
@date:   01/2013

@ [oliva's lab](http://sbi.imim.es)

@class: StorableObject
'''
from abc import ABCMeta
try:
    import cPickle as pickle
except:
    import pickle

from .  import File
from .. import SBIglobals as SBIg


class StorableObject(object):
    '''
    An abstract "dumping" class.

    This means that it is basically useful for those who would like to extend
    this library.
    Basically, it gives the object the ability to be "dumped" on disk and be
    recovered afterwards.

    '''
    __metaclass__ = ABCMeta

    def dump(self, object_file, overwrite = None):
        '''
        Stores the object into a file.

        @param:    object_file
        @pdef:     Name for the output file
        @ptype:    {String}

        @param:    overwrite
        @pdef:     write over a existing file
        @pdefault: _SBIglobals.overwrite_
        @ptype:    {Boolean}
        '''
        SBIg.alert('verbose', self, 'Writting object to file {0}'.format(object_file))
        dumpFile = File(file_name = object_file, action = 'w', overwrite = overwrite)
        pickle.dump(self, dumpFile._fd)
        dumpFile.close()

    @staticmethod
    def load(object_file):
        '''
        Retrieves the object from a python object file

        @param:    object_file
        @pdef:     Name of the file containing the object
        @ptype:    {String}

        @return: child of {StorableObject}
        '''
        SBIg.alert('verbose', StorableObject(), 'Preparing to load object from file {0}'.format(object_file))
        Object   = None
        loadFile = File(file_name = object_file, action='r')
        Object   = pickle.load(loadFile._fd)
        loadFile.close()
        return Object



"""
AminoAcid

author: jbonet
date:   02/2013

@oliva's lab
"""

from .          import Residue,        ResidueOfNucleotide
from SBI.data   import aminoacids3to1, aminoacids1to3, aminoacids_polarity_boolean, aminoacids_surface
from SBI        import SBIglobals

import numpy         as np
import scipy.spatial as sp

class ResidueOfAminoAcid(Residue):
    """
    A {AminoAcid} collects a series of {AminoAtoms}
    """
    def __init__(self, number, version, Rtype, mode):
        """
        @type  number: Integer
        @param number: Residue number

        @type  version: Char
        @param version: Optional char used on pdbs to change count

        @type  type: String
        @param type: Residue type

        @type  mode: String
        @param mode: Residue mode: ATOM or HETATM
        """
        super(ResidueOfAminoAcid, self).__init__(number = number, version = version, Rtype = Rtype, mode = mode)

        self._ca       = None
        self._cb       = None
        self._n        = None
        self._c        = None
        self._o        = None
        
        self._dssp     = None

    #
    # ATTRIBUTES
    #
    @property
    def ca(self):
    	"""
    	Returns the C-alpha {AminoAtom}
    	@rtype: {AminoAtom}
    	"""
    	return self._ca

    @property
    def cb(self):
    	"""
    	Returns the C-beta {AminoAtom}
    	@rtype: {AminoAtom}
    	"""
    	return self._cb

    @property
    def n(self):
        return self._n

    @property
    def c(self):
        return self._c

    @property
    def o(self):
        return self._o

    @property
    def backbone(self):
    	"""
    	Returns a list of the backbone {AminoAtom}s
    	@rtype: List of {AminoAtom}s
    	"""
    	return self._backbone_atoms

    @property
    def single_letter(self):
    	"""
    	Returns the AminoAcid identifier as a single letter code
    	@rtype: String
    	"""
    	return aminoacids3to1[self.type]

    @property
    def standard_type(self):
    	"""
    	For some AminoAcid are HETATM, we can return the standard AminoAcid
    	@rtype: String
    	"""
    	return aminoacids1to3[self.single_letter]

    @property
    def dssp(self):
        """
        @rtype: {DSSP}
        """
        return self._dssp

    @dssp.setter
    def dssp(self, value):
        """
        Sets {DSSP}
        """
        self._dssp = value

    @property
    def secondary_structure(self):
        if self._dssp is None:
            raise AttributeError("To call secondary structure DSSP needs to be calculated")

        return self._dssp.secondary_structure

    @property
    def accessibility(self):
        if self._dssp is None:
            raise AttributeError("To call accessibility DSSP needs to be calculated")

        return self._dssp.accessibility

    @property
    def accessibility10(self):
        if self._dssp is None:
            raise AttributeError("To call accessibility DSSP needs to be calculated")

        return self._dssp.accessibility10

    @property
    def accessibilitycoded(self):
        if self._dssp is None:
            raise AttributeError("To call accessibility DSSP needs to be calculated")

        return self._dssp.accesscode

    @property
    def exposed(self):
        if self._dssp is None:
            raise AttributeError("To call exposition DSSP needs to be calculated")

        return self._dssp.exposed

    @property
    def exposed_text(self):
        if self.exposed: return 'E'
        else:            return 'B'

    @property
    def surface_area(self):
        return aminoacids_surface[self.single_letter]

    #
    # BOOLEANS
    #
    @property
    def has_ca(self):
        """
        Checks if the AminoAcid has C-alpha
        @rtype: Boolean
        """
        return self._ca is not None

    @property
    def has_cb(self):
        """
        Checks if the AminoAcid has C-beta
        @rtype: Boolean
        """
        return self._cb is not None

    @property
    def has_full_backbone(self):
        return self._ca is not None and self._c is not None \
           and self._n  is not None and self._o is not None

    @property
    def is_only_ca(self):
    	"""
    	Some times the structure only contains the C-alpha
    	@rtype: Boolean
    	"""
    	return self.has_ca and len(self) == 1

    @property
    def is_only_backbone(self):
    	"""
    	The structure does not contain side chain
    	@rtype: Boolean
    	"""
    	return len(self.backbone) == len(self)

    @property
    def is_polar(self):
        return aminoacids_polarity_boolean[self.single_letter]

    #
    # METHODS
    #
    def normalize(self):
        self._type = self.standard_type
        self._mode = 'ATOM'
        
    def remove_side_chain(self):
        # self._atoms = self._backbone_atoms
        self._sidechain_atoms = []
        self._sidechain_coordinates = None

    def _distance_ca(self, aminoacid):
        """
        Calculates the ca-ca distance between two residues

        @type  aminoacid: {AminoAcid}
        @param aminoacid: AminoAcid to which we want to find the distance

        @rtype: {Contact}
        @rtype: None if one of them does not have ca
        """
        if not self.has_ca or not aminoacid.has_ca:
            # return self._distance_min(residue = aminoacid)
            return (None, None, -1)

        return (self.ca, aminoacid.ca, self.ca.distance(atom = aminoacid.ca))


    def _distance_cb(self, aminoacid):
        """
        Calculates the cb-cb distance between two residues.
        If one does not has cb, reverts to ca-cb or cb-ca.
        If none have cb, reverts to ca-ca.

        @type  aminoacid: {AminoAcid}
        @param aminoacid: AminoAcid to which we want to find the distance

        @rtype: {Contact}
        @rtype: None if one of them does not have ca
        """
        if (not self.has_ca and not self.has_cb) or (not aminoacid.has_ca and not aminoacid.has_cb):
            # return (self._distance_min(residue = aminoacid))
            return (None, None, -1)
        if not self.has_cb and not aminoacid.has_cb:
            # return self._distance_ca(aminoacid = aminoacid)
            return (None, None, -1)
        elif not self.has_cb:
            return (self.ca, aminoacid.cb, self.ca.distance(atom = aminoacid.cb))
        elif not aminoacid.has_cb:
            return (self.cb, aminoacid.ca, self.cb.distance(atom = aminoacid.ca))
        else:
            return (self.cb, aminoacid.cb, self.cb.distance(atom = aminoacid.cb))

    def _distance_cb_backbone(self, nucleotide):
        if self.has_cb:
            cb_coord = np.array(np.zeros(3))
            cb_coord = np.vstack((cb_coord, self.cb.coordinates))
            cb_coord = np.delete(cb_coord, 0, 0)
            atom     = self.cb
        elif self.has_ca:
            cb_coord = np.array(np.zeros(3))
            cb_coord = np.vstack((cb_coord, self.ca.coordinates))
            cb_coord = np.delete(cb_coord, 0, 0)
            atom     = self.ca
        else:
            return (None, None, -1)
        if nucleotide._backbone_coordinates is not None:
            n_coord = nucleotide._backbone_coordinates
            n_atoms = nucleotide._backbone_atoms
        else:
            n_coord = nucleotide._sidechain_coordinates
            n_atoms = nucleotide._sidechain_atoms
        SBIglobals.alert('deepdebug', self, '\tAminoAcid coordenate {0}'.format(cb_coord))
        SBIglobals.alert('deepdebug', self, '\tNucleotide coordenates {0}'.format(n_coord))
        distances = sp.distance.cdist(cb_coord, n_coord)
        index     = np.unravel_index(distances.argmin(), distances.shape)
        return (atom, n_atoms[index[1]], distances.min())

    #
    # OVERRIDE PARENT'S METHODS
    #
    def add_atom(self, atom):
        """
        Adds a new {AminoAtom} to the {AminoAcid}
        This includes:
        	filling the all_coordinates attribute (parent)
        	assign CA, CB and backbone
        @type  atom: {AminoAtom}
        @param atom: New {AminoAtom} added to the {AminoAcid}
        """
        super(ResidueOfAminoAcid, self).add_atom(atom = atom)

        if atom.is_Calpha:
        	self._ca = atom
        elif atom.is_Cbeta:
        	self._cb = atom
        elif atom.is_C:
            self._c  = atom
        elif atom.is_N:
            self._n  = atom
        elif atom.is_O:
            self._o  = atom

    def distance(self, residue, dist_type):
        """
        Calculates the distance between two residues

        @type  residue: {AminoAcid}
        @param residue: AminoAcid to which we want to find the distance

        @type  dist_type: String
        @param dist_type: Type of distance to evaluate
        @limit dist_type: Accepted: distance.types (ca, cb, min, geometric)

        @rtype: (AT1, AT2, DISTANCE)
        """
        if not isinstance(residue, ResidueOfAminoAcid):
            if isinstance(residue, ResidueOfNucleotide):
                if dist_type.strip().lower() != 'cbbackbone':
                    return super(ResidueOfAminoAcid, self).distance(residue = residue, dist_type = dist_type)
                else:
                    return self._distance_cb_backbone(nucleotide = residue)
            else:
                return super(ResidueOfAminoAcid, self).distance(residue = residue, dist_type = dist_type)
        if dist_type.strip().lower() == 'ca':
            return self._distance_ca(aminoacid = residue)
        if dist_type.strip().lower() == 'cb':
            return self._distance_cb(aminoacid = residue)
        else:
            return super(ResidueOfAminoAcid, self).distance(residue = residue, dist_type = dist_type)

    def follows(self, residue):
        return residue.is_followed(self)

    def is_followed(self, residue):
        c0, ca0, n0 = self.c,    self.ca,    self.n
        c1, ca1, n1 = residue.c, residue.ca, residue.n

        if c0  is not None and n1  is not None: return c0.distance(n1)   <= 1.5 # measured distance is 1.3 
        if ca0 is not None and ca1 is not None: return ca0.distance(ca1) <= 4   # measured distance around 3.8
        if c0  is not None and c1  is not None: return c0.distance(c1)   <= 4
        if n0  is not None and n1  is not None: return n0.distance(n1)   <= 4

    #
    # OVERRIDE DEFAULT OBJECT METHODS
    #
    def __repr__(self):
        repre = []
        if self._dssp is None:
            repre.append('<{0.__class__.__name__}: [{0.type}, {0.number}, {0.version}]>'.format(self))
        else:
            repre.append('<{0.__class__.__name__}: [{0.type}, {0.number}, {0.version}, ({0.secondary_structure}, {0.exposed_text})]>'.format(self))
        for atom in self.atoms:
            repre.append("\t" + repr(atom))
        return "\n".join(repre)


"""drugBanklink

author: jbonet
date:   02/2014

@oliva's lab
"""

"""
Import Standard Libraries
"""
import os, re
import urllib

"""
Dependences in SBI library
"""
from SBI.databases import drugBankftp
from SBI.beans.Path     import Path
from SBI.beans.File     import File

class DrugBanklink(object):
    def __init__(self, local = None):
        self._local    = os.path.abspath(local)
        self._tagets   = None
        self._main     = None
        self._drugfile = None
        if local is not None:
            self.local = local

    """ATTRIBUTES"""
    @property
    def local(self):         return self._local
    @local.setter
    def local(self, value):
        self._local    = os.path.abspath(value)
        self._target   = os.path.join(self._local, os.path.split(drugBankftp['targets'])[-1])
        self._main     = os.path.join(self._local, os.path.split(drugBankftp['main'])[-1])
        self._drugfile = os.path.join(self._local, 'drugbank.gz')

    @property
    def source(self):
        return drugBankftp['show']

    """BOOLEANS"""
    @property
    def has_local(self):    return self._local is not None

    """METHODS"""
    def download(self):
        if not self.has_local:
            raise NameError('A local drugBank database directory must be defined.')

        Path.mkdir(self.local)
        urllib.urlretrieve(drugBankftp['targets'], self._target)
        urllib.urlretrieve(drugBankftp['main'],    self._main)

        self._process()

        return True

    @property
    def localDrugs(self):
        drgFile = File(self._drugfile, 'r')
        for drg_line in drgFile.descriptor:
            yield drg_line

    """PRIVATE METHODS"""
    def _process(self):

        targets = self._process_targets()
        drugs   = self._process_drugs(targets)

        drugFile = File(self._drugfile,'w', True)
        for d in drugs:
            drugFile.write(repr(d)+"\n")
        drugFile.close()

    def _process_targets(self):
        import csv, StringIO, zipfile

        targets = {}
        z       = zipfile.ZipFile(self._target, 'r')
        data    = StringIO.StringIO(z.read(z.namelist()[0]))
        reader  = csv.reader(data)
        for row in reader:
            if row[5] != '':
                targets[row[0]] = (row[5], row[6])
        z.close()
        return targets

    def _process_drugs(self, targets):
        import zipfile

        drugs = []
        z = zipfile.ZipFile(self._main, 'r')
        for line in z.open(z.namelist()[0]):
            if line.startswith('  <drug'):
                drugs.append(Drug())
                drug        = drugs[-1]
                drug.type   = line
            elif line.startswith('    <drugbank-id'):                   drug.dbid       = line
            elif line.startswith('    <name'):                          drug.name       = line
            elif line.startswith('    <description>'):                  drug.desc       = line
            elif line.startswith('    <cas-number'):                    drug.cas        = line
            elif line.startswith('      <secondary-accession-number'):  drug.secondary_accession_numbers = line
            elif line.startswith('      <group'):                       drug.groups     = line
            elif line.startswith('        <kind'):                      propkind        = line
            elif line.startswith('        <value'):                     drug.dproperty  = (propkind, line)
            elif line.startswith('      <category'):                    drug.categories = line
            elif line.startswith('        <resource>'):                 propkind        = line
            elif line.startswith('        <identifier>'):               drug.external   = (propkind, line)
            elif line.startswith('      <synonym>'):                    drug.synonym    = line
            elif 'partner=' in line:
                m = re.search('<(\w+) partner="(\d+)"', line)
                if m.group(2) in targets:
                    drug.partners = [m.group(1), targets[m.group(2)]]
                else:
                    drug.partners = [m.group(1), ('','')]
                partner = drug.partners[-1]
            elif line.startswith('          <action>'):
                partner.append(re.search(">([\s\S]+)</", line).group(1))
            elif line.startswith('        <known-action>'):
                partner.append(re.search(">(\S+)</", line).group(1).lower())
            # elif line.startswith('  </drug'):
            #     # print line.strip()
            #     print drug
        z.close()
        return drugs

class Drug(object):

    properties_of_interest = ["Hydrophobicity", "Isoelectric Point", "Molecular Weight", "Molecular Formula"]

    def __init__(self, inline = None):
        if inline is not None: inline = inline.strip().split('\t')

        self._type = None if inline is None else inline[0]
        self._dbid = None if inline is None else inline[1]
        self._name = None if inline is None else inline[2]
        self._cas  = None if inline is None else inline[3]
        self._desc = None if inline is None else inline[4]
        self._san  = []   if inline is None else eval(inline[5])
        self._grps = []   if inline is None else eval(inline[6])
        self._prop = {}   if inline is None else eval(inline[7])
        self._cate = []   if inline is None else eval(inline[8])
        self._extr = []   if inline is None else eval(inline[9])
        self._synm = []   if inline is None else eval(inline[10])
        self._part = []   if inline is None else eval(inline[11])

    @property
    def type(self): return self._type
    @type.setter
    def type(self, value): self._type = re.search('type\=\"(\w+\s*\w+)\"', value).group(1)

    @property
    def dbid(self): return self._dbid
    @dbid.setter
    def dbid(self, value): self._dbid = re.search(">(DB\d+)</", value).group(1)

    @property
    def name(self): return self._name
    @name.setter
    def name(self, value): self._name = re.search(">([\S\s]+)</", value).group(1)

    @property
    def cas(self): return self._cas
    @cas.setter
    def cas(self, value):
        m = re.search(">([\d\-]+)</", value)
        self._cas = m.group(1) if m else ''

    @property
    def desc(self): return self._desc
    @desc.setter
    def desc(self, value):
        m = re.search(">([\S\s]+)\s*[<|&]", value)
        self._desc = m.group(1).rstrip('< ') if m else ''

    @property
    def secondary_accession_numbers(self): return self._san
    @secondary_accession_numbers.setter
    def secondary_accession_numbers(self, value): self._san.append(re.search(">([\S\s]+)</", value).group(1))

    @property
    def groups(self): return self._grps
    @groups.setter
    def groups(self, value): self._grps.append(re.search(">([\S\s]+)</", value).group(1))

    @property
    def dproperty(self): return self._prop
    @dproperty.setter
    def dproperty(self, value):
        if re.search(">([\S\s]+)</", value[0]).group(1) in self.properties_of_interest:
            self._prop[re.search(">([\S\s]+)</", value[0]).group(1)] = re.search(">([\S\s]+)</", value[1]).group(1).split()[0]

    @property
    def categories(self): return self._cate
    @categories.setter
    def categories(self, value): self._cate.append(re.search(">([\S\s]+)</", value).group(1))

    @property
    def external(self): return self._extr
    @external.setter
    def external(self, value):
        self._extr.append(tuple([re.search(">([\S\s]+)</", value[0]).group(1), re.search(">([\S\s]+)</", value[1]).group(1)]))

    @property
    def synonym(self): return self._synm
    @synonym.setter
    def synonym(self, value): self._synm.append(re.search(">([\S\s]+)</", value).group(1))

    @property
    def partners(self): return self._part
    @partners.setter
    def partners(self, value): self._part.append(value)

    def __repr__(self):
        data = []
        data.append(self._type)
        data.append(self._dbid)
        data.append(self._name)
        data.append(self._cas)
        data.append(self._desc)
        data.append(repr(self._san))
        data.append(repr(self._grps))
        data.append(repr(self._prop))
        data.append(repr(self._cate))
        data.append(repr(self._extr))
        data.append(repr(self._synm))
        data.append(repr(self._part))

        return "\t".join(data)



import sys
import argparse
import os
import re
from Bio import SeqIO as SeqIE
from Bio import ExPASy
from Bio.Seq import Seq
from Bio import SeqRecord
from Bio.Alphabet import IUPAC
import ConfigParser
import itertools
import shutil
import subprocess
import time
import random
import numpy as np


# Add '.' to sys.path
src_path = os.path.abspath(os.path.dirname(__file__))
sys.path.append(src_path)
script_path = src_path+"/../scripts"
sys.path.append(script_path)

# Read configuration file
config = ConfigParser.ConfigParser()
config_file = os.path.join(script_path, 'config.ini')
config.read(config_file)

# Add SBI library path to sys.path
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','sbi_library_path'))
sys.path.append(src_path)

# Add functions path to sys.path    
src_path = os.path.join(config.get('Paths','modppi_path'),config.get('Paths','functions_path'))
sys.path.append(src_path)

# Import functions
import functions
from SeqIO import *
from functions import *
from SBI.external.blast import blast_parser
from SBI.structure.contacts import Complex
from SBI.sequence import Sequence
from SBI.structure import PDB

#Parameters
#Original Boltzman constant (kcal/molK):1.9872036e-3
kBo  = 1.9872036e-3
#Temperature
T    = 300.0
#Least Square lineal fitting using Affinity Benchmark
#Maff = 0.115
#Caff = -7.326
#Maff = 1.0
Maff = 0.0003


#Boltzman constant fixed
kB   = kBo/Maff



def fileExist (file):
 return os.path.exists(file) and os.path.isfile(file)

def distribution(E):
  mean=0.0
  sigma=0.0
  nn=len([x for x in E.itervalues() if x<0])
  n=len([x for x in E.itervalues()])
  minimum=0.0
  maximum=0.0
  if n>1: minimum=min([float(x) for x in E.itervalues()])
  if n>1: maximum=max([float(x) for x in E.itervalues()])
  if n<2 : return (mean,sigma,minimum,maximum)
  for m,ee in E.iteritems():
    if ee<0:
      mean =mean +float(ee)/nn
      sigma=sigma+float(ee)*float(ee)/n
  sigma = sigma - mean*mean
  if sigma > 0: 
     sigma = np.sqrt(sigma)
     return (mean,sigma,minimum,maximum)
  else:
     return (mean,0.0,minimum,maximum)

  
def partition_function(E):
 Z  = 0.0
 (mean,sigma,minimum,maximum)=distribution(E)
 p=minimum    
 #p=min([minimum+sigma,mean-3*sigma,minimum-(minimum/2)])    
 for m,ee in E.iteritems():
   e=(float(ee) - p)/(kB*T)
   try:
     if e > 1.0e+6: print "Out of limit Model %s is %f > 1.0e+6 \n"%(m,e)
     if e < 1.0e+6 :  Z = Z + np.exp(-e)
     if e >= 1.0e+6:  Z = Z 
     if e<0: print "Out of limit Model %s is  %f < 0\n"%(m,e)
   except  Exception as err:
      sys.stderr.write("ERROR: %s\n"%err)
      sys.stderr.write("ERROR: %s %s\n"%(m,str(e)))
      continue
 return (Z,p)

def average_ensemble (S,E):
  (Z,p)= partition_function( E )
  a    = 0.0
  for m,ee in E.iteritems():
      e=(float(ee) - p)/(kB*T)
      if S.has_key(m):
         try:
           if e < 1.0e+6:  a = a + float(S.get(m)) * np.exp( -e ) / Z
           if e >= 1.0e+6: a = a
         except  Exception as err:
           sys.stderr.write("ERROR: %s\n"%err)
           sys.stderr.write("ERROR: %s %s\n"%(m,str(e)))
           continue
  return a

def FEP (A,B):
  (ZA,p) = partition_function( A )
  x      = 0.0
  data   = [ (float(B.get(m)) - float(A.get(m))) for m,ee in A.iteritems() if B.has_key(m) ]
  if len(data) > 0: 
     d_min  = min(data)
  else:
     return 0.0
  for m,ee in A.iteritems():
    e=(float(ee) - p)/(kB*T)
    if B.has_key(m):
       try:
         d = float(B.get(m)) - float(A.get(m)) - d_min 
         if e < 1.0e+6 and d/(kB*T)< 1.0e+6 :  x = x + np.exp( -d/(kB*T) ) * np.exp(-e)/ZA
         if e >= 1.0e+6 or d/(kB*T)>=1.0e+6 :  x = x
       except  Exception as err:
         sys.stderr.write("ERROR: %s\n"%err)
         sys.stderr.write("ERROR: %s %s\n"%(m,str(e)))
         continue
  if x>0:
     fep = - kB * T * np.log(x) + d_min
  else:
     fep = d_min
  return fep
  

if len(sys.argv)==1:
 sys.stdout.write("   Execution is 'Exponential_Averaging_FEP.py <directory>    <variable>   <energy for Z>' \n \
 -- where directory is the output folder of modelist.py,  \n \
 -- 'energy for Z' is the score to calculate Z using Zwanzig equation (ddG_mean by default), and \n \
 -- 'variable' is the function to evaluate the comparison of exponential averages (ddG_mean by default)\n")
 exit(0)

input_dir=sys.argv[1]

if len(sys.argv)>3: 
  energy=sys.argv[3]
  score=sys.argv[2]
elif len(sys.argv)==3:
  energy="ddG_mean"
  score=sys.argv[2]
else: 
  energy="ddG_mean"
  score="ddG_mean"



files=os.listdir(input_dir)

ppi_sets=set()
for input_file in files:
   if input_file.endswith(".ppi"): 
      ppi_sets.add(input_file)

global_output = os.path.join(input_dir,"FEP_"+score+"_Z_with_"+energy+".out")
fg = open(global_output,"w")
fg.write("%5s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%s\n"%("#ppi","FEP direct","FEP invers","<FEP>","-kTln(Z'/Z)","score WT","score MUT","Diff.score","PPI models"))
for file_ppi in ppi_sets:
    output = os.path.join(input_dir,file_ppi+"_"+score+"_Z_with_"+energy+".out")
    print "Write Output in %s"%(output)
    fo = open(output,"w")
    fo.write("%5s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%s\n"%("#ppi","FEP direct","FEP invers","<FEP>","-kTln(Z'/Z)","score WT","score MUT","Diff.score","PPI models"))
    ppis=[]
    fd=open(os.path.join(input_dir,file_ppi),'r')
    for line in fd:
      pair=line.split()
      ppis.append((pair[0],pair[1]))
    fd.close()
    root_ppi=file_ppi.rstrip(".ppi")
    data=root_ppi.split("_")
    label=data[0]
    wt=(data[1],data[-1])
    wt_file=data[1]+"::"+data[-1]
    if not os.path.exists(os.path.join(input_dir,"models",wt_file)): continue
    print "Check files in %s"%(wt_file)
    files_wt = os.listdir(os.path.join(input_dir,"models",wt_file))
    files_ddg_wt_path=set()
    files_scr_wt_path=set()
    number_of_clusters=set()
    for f in files_wt:
      if f.endswith("out"):
        cluster_group = f.split("cluster_")
        if len(cluster_group)>0:
           number_group=cluster_group[1].split(".list.")
           if len(number_group)>0:
              if number_group[0] != "0":
                 number_of_clusters.add(number_group[0])
    #print "Clusters %s"%number_of_clusters
    for number_group in number_of_clusters:
           ddg_wt_file=root_ppi+"_"+wt_file+"_cluster_"+number_group+".list."+energy+"."+root_ppi+".out"
           scr_wt_file=root_ppi+"_"+wt_file+"_cluster_"+number_group+".list."+score +"."+root_ppi+".out"
           files_ddg_wt_path.add(os.path.join(input_dir,"models",wt_file,ddg_wt_file))
           files_scr_wt_path.add(os.path.join(input_dir,"models",wt_file,scr_wt_file))
    E_wt={}
    S_wt={}
    for ddg_wt_path in files_ddg_wt_path:
       if not fileExist(ddg_wt_path): continue
       print "Open %s"%ddg_wt_path
       fd=open(ddg_wt_path,"r")
       for line in fd:
         model_file,ene = line.split()
         if model_file.endswith(".h") or model_file.endswith(".h_AB") or model_file.endswith(".h_BA"):
            model = os.path.basename(model_file)
            E_wt.setdefault(model,float(ene))
       fd.close()
    for scr_wt_path in files_scr_wt_path:
       if not fileExist(scr_wt_path): continue
       print "Open %s"%scr_wt_path
       fd=open(scr_wt_path,"r")
       for line in fd:
         model_file,scr = line.split()
         if model_file.endswith(".h") or model_file.endswith(".h_AB") or model_file.endswith(".h_BA"):
            model = os.path.basename(model_file)
            S_wt.setdefault(model,float(scr))
       fd.close()
    if len(files_ddg_wt_path)>0 and len(files_scr_wt_path)>0:
       n = 0;
       for ppi in ppis:
           type_file     = ppi[0]+"::"+ppi[1]
           if not os.path.exists(os.path.join(input_dir,"models",type_file)): continue
           print "Check files in %s"%(type_file)
           files_type = os.listdir(os.path.join(input_dir,"models",type_file))
           files_ddg_type_path=set()
           files_scr_type_path=set()
           number_of_clusters=set()
           for f in files_type:
               if f.endswith("out"):
                 cluster_group = f.split("cluster_")
                 if len(cluster_group)>0:
                    number_group=cluster_group[1].split(".list.")
                    if len(number_group)>0:
                       if number_group[0] != "0":
                          number_of_clusters.add(number_group[0])
           #print "Clusters %s"%number_of_clusters
           for number_group in number_of_clusters:
                    ddg_type_file=root_ppi+"_"+type_file+"_cluster_"+number_group+".list."+energy+"."+root_ppi+".out"
                    scr_type_file=root_ppi+"_"+type_file+"_cluster_"+number_group+".list."+score +"."+root_ppi+".out"
                    files_ddg_type_path.add(os.path.join(input_dir,"models",type_file,ddg_type_file))
                    files_scr_type_path.add(os.path.join(input_dir,"models",type_file,scr_type_file))
           E_type={}
           S_type={}
           for ddg_type_path in files_ddg_type_path:
               if not fileExist(ddg_type_path): continue
               print "Open %s"%ddg_type_path
               fd=open(ddg_type_path,"r")
               for line in fd:
                 model_file,ene = line.split()
                 if model_file.endswith(".h") or model_file.endswith(".h_AB") or model_file.endswith(".h_BA"):
                    model = os.path.basename(model_file)
                    E_type.setdefault(model,float(ene))
               fd.close()
           for scr_type_path in files_scr_type_path:
               if not fileExist(scr_type_path):continue
               print "Open %s"%scr_type_path
               fd=open(scr_type_path,"r")
               for line in fd:
                 model_file,scr = line.split()
                 if model_file.endswith(".h") or model_file.endswith(".h_AB") or model_file.endswith(".h_BA") :
                    model = os.path.basename(model_file)
                    S_type.setdefault(model,float(scr))
               fd.close()
           if len(files_ddg_type_path)>0 and len(files_scr_type_path)>0:
               n = n + 1
               Scr_wt       = average_ensemble( S_wt,E_wt )
               Scr_type     = average_ensemble( S_type,E_type )
               delta_scr    = Scr_type - Scr_wt
               FEP_direct   = FEP(E_wt,E_type)
               FEP_invers   = FEP(E_type,E_wt)
               FEP_average  = (FEP_direct - FEP_invers)/2
               (ZA,pA)      = partition_function( E_wt )
               (ZB,pB)      = partition_function( E_type )
               if ZA>0 and ZB>0: partition = - kB * T * np.log(ZB/ZA)
               fo.write("%5d\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%s\n"%(n,FEP_direct,FEP_invers,FEP_average,partition,Scr_wt,Scr_type,delta_scr,type_file))
               fg.write("%5d\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%s\n"%(n,FEP_direct,FEP_invers,FEP_average,partition,Scr_wt,Scr_type,delta_scr,type_file))
               print "Add %5d\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%10.5e\t%s\n"%(n,FEP_direct,FEP_invers,FEP_average,partition,ZA,ZB,Scr_wt,Scr_type,delta_scr,type_file)
    fo.close()
fg.close()         
print "Done"

    

